"""
famous for its use of separable convolutions,
usually the SOTA image classifier,
however seems not have been used in physiological signal processing tasks
"""

import textwrap
from copy import deepcopy
from itertools import repeat
from numbers import Real
from typing import Optional, Sequence, Union, List

from torch import Tensor, nn

from ...cfg import CFG
from ...models._nets import DownSample, MultiConv
from ...utils.misc import add_docstring, CitationMixin
from ...utils.utils_nn import (
    SizeMixin,
    compute_sequential_output_shape,
    compute_sequential_output_shape_docstring,
)


__all__ = [
    "Xception",
    "XceptionEntryFlow",
    "XceptionMiddleFlow",
    "XceptionExitFlow",
    "XceptionMultiConv",
]


if not hasattr(nn, "Dropout1d"):
    nn.Dropout1d = nn.Dropout  # added in pytorch 1.12


_DEFAULT_CONV_CONFIGS = CFG(
    ordering="acb",
    conv_type="separable",
    batch_norm=True,
    subsample_mode="max",
    activation="relu",
    kw_activation={"inplace": True},
    kernel_initializer="he_normal",
    kw_initializer={},
)


class XceptionMultiConv(nn.Module, SizeMixin, CitationMixin):
    """Xception multi-convolutional block.

    -> n(2 or 3) x (activation -> norm -> sep_conv) (-> optional sub-sample) ->
    |-------------------------------- shortcut ------------------------------|

    Parameters
    ----------
    in_channels : int
        Number of channels in the input tensor.
    num_filters : Sequence[int]
        Number of channels produced by the main stream convolutions.
        The length of `num_filters` indicates the number of convolutions.
    filter_lengths : int or Sequence[int]
        Length(s) of the filters (kernel size).
    subsample_length : int, default 1
        Stride of the main stream subsample layer.
    subsample_kernel : int, optional
        Kernel size of the main stream subsample layer.
        If not set, defaults to `subsample_length`.
    dilations : int or Sequence[int], default 1
        Dilation(s) of the convolutions.
    groups : int, default 1
        Connection pattern (of channels) of the inputs and outputs.
    dropouts : float or dict or Sequence[float] or Sequence[dict], default 0.0
        If is dictionary, it should contain the keys ``"p"`` and ``"type"``,
        where ``"p"`` is the dropout rate and ``"type"`` is the type of dropout,
        which can be either ``"1d"`` (:class:`torch.nn.Dropout1d`) or
        ``None`` (:class:`torch.nn.Dropout`).
    config : dict, optional
        Other parameters, including
        activation choices, weight initializer, batch normalization choices, etc.,
        for the convolutional layers,
        and subsampling modes for subsampling layers, etc.

    """

    __name__ = "XceptionMultiConv"

    def __init__(
        self,
        in_channels: int,
        num_filters: Sequence[int],
        filter_lengths: Union[Sequence[int], int],
        subsample_length: int = 1,
        subsample_kernel: Optional[int] = None,
        dilations: Union[Sequence[int], int] = 1,
        groups: int = 1,
        dropouts: Union[Sequence[float], float] = 0.0,
        **config,
    ) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.__num_filters = list(num_filters)
        self.__num_convs = len(self.__num_filters)
        if isinstance(filter_lengths, int):
            self.__filter_lengths = list(repeat(filter_lengths, self.__num_convs))
        else:
            self.__filter_lengths = list(filter_lengths)
        assert self.__num_convs == len(self.__filter_lengths), (
            f"the main stream has {self.__num_convs} convolutions, "
            f"while `filter_lengths` indicates {len(self.__filter_lengths)}"
        )
        self.__subsample_length = subsample_length
        self.__subsample_kernel = subsample_kernel or subsample_length
        self.__groups = groups
        if isinstance(dilations, int):
            self.__dilations = list(repeat(dilations, self.__num_convs))
        else:
            self.__dilations = list(dilations)
        assert self.__num_convs == len(self.__dilations), (
            f"the main stream has {self.__num_convs} convolutions, "
            f"while `dilations` indicates {len(self.__dilations)}"
        )
        if isinstance(dropouts, (Real, dict)):
            self.__dropouts = list(repeat(dropouts, self.__num_convs))
        else:
            self.__dropouts = list(dropouts)
        assert self.__num_convs == len(self.__dropouts), (
            f"the main stream has {self.__num_convs} convolutions, "
            f"while `dropouts` indicates {len(self.__dropouts)}"
        )
        self.config = CFG(deepcopy(_DEFAULT_CONV_CONFIGS))
        self.config.update(deepcopy(config))

        self.main_stream_conv = MultiConv(
            in_channels=self.__in_channels,
            out_channels=self.__num_filters,
            filter_lengths=self.__filter_lengths,
            subsample_lengths=1,
            dilations=self.__dilations,
            groups=self.__groups,
            dropouts=self.__dropouts,
            **self.config,
        )
        if self.__subsample_length > 1:
            self.subsample = DownSample(
                down_scale=self.__subsample_length,
                in_channels=self.__num_filters[-1],
                kernel_size=self.__subsample_kernel,
                groups=self.__groups,
                padding=(self.__subsample_kernel - 1) // 2,
                mode=self.config.subsample_mode,
            )
            self.shortcut = DownSample(
                down_scale=self.__subsample_length,
                in_channels=self.__in_channels,
                out_channels=self.__num_filters[-1],
                groups=self.__groups,
                kernel_size=1,
                norm=self.config.get("norm", self.config.get("batch_norm")),
                mode="conv",
            )
        else:
            self.subsample = nn.Identity()
            self.shortcut = nn.Identity()

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        Returns
        -------
        output : torch.Tensor
            Output tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        """
        main_out = self.main_stream_conv(input)
        main_out = self.subsample(main_out)
        residue = input
        residue = self.shortcut(residue)
        output = residue + main_out
        return output

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the module.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input tensor.
        batch_size : int, optional
            Batch size of the input tensor.

        Returns
        -------
        output_shape : sequence
            The output shape of the module.

        """
        output_shape = self.main_stream_conv.compute_output_shape(seq_len, batch_size)
        if not isinstance(self.subsample, nn.Identity):
            output_shape = self.subsample.compute_output_shape(
                output_shape[-1], output_shape[0]
            )
        return output_shape


class XceptionEntryFlow(nn.Sequential, SizeMixin):
    """Entry flow of the Xception model.

    The entry flow is consisting of
    2 initial convolutions which subsamples at the first one,
    followed by several Xception blocks of 2 convolutions
    and of sub-sampling size 2.

    Parameters
    ----------
    in_channels : int
        Number of channels in the input signal.
    init_num_filters : Sequence[int]
        Number of filters (output channels) of the initial convolutions.
    init_filter_lengths : int or Sequence[int]
        Filter length(s) (kernel size(s)) of the initial convolutions.
    init_subsample_lengths : int or Sequence[int]
        Subsampling length(s) (stride(s)) of the initial convolutions.
    num_filters : Sequence[int] or Sequence[Sequence[int]]
        Number of filters of the convolutions of Xception blocks.
    filter_lengths : int or Sequence[int] or Sequence[Sequence[int]]
        Filter length(s) of the convolutions of Xception blocks.
    subsample_lengths : int or Sequence[int]
        Subsampling length(s) of the Xception blocks.
    subsample_kernels : int or Sequence[int], optional
        Subsampling kernel size(s) of the Xception blocks.
    dilations : int or Sequence[int] or Sequence[Sequence[int]], default 1
        Dilation(s) of the convolutions of Xception blocks.
    groups : int, default 1
        Connection pattern (of channels) of the inputs and outputs.
    dropouts : float or dict or Sequence[Union[float, dict]] or Sequence[Sequence[Union[float, dict]]], default 0.0
        Dropout(s) after each :class:`Conv_Bn_Activation` blocks
        in the Xception blocks.
    block_dropouts : float or dict or Sequence[Union[float, dict]], default 0.0
        Dropout(s) after the Xception blocks.
    config : dict, optional
        Other parameters, Xception blocks and initial convolutions, including
        activation choices, weight initializer, batch normalization choices, etc.
        for the convolutional layers,
        and subsampling modes for subsampling layers, etc.

    """

    __name__ = "XceptionEntryFlow"

    def __init__(
        self,
        in_channels: int,
        init_num_filters: Sequence[int],
        init_filter_lengths: Union[int, Sequence[int]],
        init_subsample_lengths: Union[int, Sequence[int]],
        num_filters: Union[Sequence[int], Sequence[Sequence[int]]],
        filter_lengths: Union[int, Sequence[int], Sequence[Sequence[int]]],
        subsample_lengths: Union[int, Sequence[int]],
        subsample_kernels: Optional[Union[int, Sequence[int]]] = None,
        dilations: Union[int, Sequence[int], Sequence[Sequence[int]]] = 1,
        groups: int = 1,
        dropouts: Union[
            float,
            dict,
            Sequence[Union[float, dict]],
            Sequence[Sequence[Union[float, dict]]],
        ] = 0.0,
        block_dropouts: Union[float, dict, Sequence[Union[float, dict]]] = 0.0,
        **config,
    ) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.__num_filters = list(num_filters)
        self.__num_blocks = len(self.__num_filters)
        if isinstance(filter_lengths, int):
            self.__filter_lengths = list(repeat(filter_lengths, self.__num_blocks))
        else:
            self.__filter_lengths = list(filter_lengths)
        assert self.__num_blocks == len(self.__filter_lengths), (
            f"the entry flow has {self.__num_blocks} blocks, "
            f"while `filter_lengths` indicates {len(self.__filter_lengths)}"
        )
        if isinstance(subsample_lengths, int):
            self.__subsample_lengths = list(
                repeat(subsample_lengths, self.__num_blocks)
            )
        else:
            self.__subsample_lengths = list(subsample_lengths)
        assert self.__num_blocks == len(self.__subsample_lengths), (
            f"the entry flow has {self.__num_blocks} blocks, "
            f"while `subsample_lengths` indicates {len(self.__subsample_lengths)}"
        )
        if subsample_kernels is None:
            self.__subsample_kernels = deepcopy(self.__subsample_lengths)
        elif isinstance(subsample_kernels, int):
            self.__subsample_kernels = list(
                repeat(subsample_kernels, self.__num_blocks)
            )
        else:
            self.__subsample_kernels = list(subsample_kernels)
        assert self.__num_blocks == len(self.__subsample_kernels), (
            f"the entry flow has {self.__num_blocks} blocks, "
            f"while `subsample_kernels` indicates {len(self.__subsample_kernels)}"
        )
        if isinstance(dilations, int):
            self.__dilations = list(repeat(dilations, self.__num_blocks))
        else:
            self.__dilations = list(dilations)
        assert self.__num_blocks == len(self.__dilations), (
            f"the entry flow has {self.__num_blocks} blocks, "
            f"while `dilations` indicates {len(self.__dilations)}"
        )
        if isinstance(dropouts, (Real, dict)):
            self.__dropouts = list(repeat(dropouts, self.__num_blocks))
        else:
            self.__dropouts = list(dropouts)
        assert self.__num_blocks == len(self.__dropouts), (
            f"the entry flow has {self.__num_blocks} blocks, "
            f"while `dropouts` indicates {len(self.__dropouts)}"
        )
        if isinstance(block_dropouts, (Real, dict)):
            self.__block_dropouts = list(repeat(block_dropouts, self.__num_blocks))
        else:
            self.__block_dropouts = list(block_dropouts)
        assert self.__num_blocks == len(self.__block_dropouts), (
            f"the entry flow has {self.__num_blocks} blocks, except the initial convolutions, "
            f"while `block_dropouts` indicates {len(self.__block_dropouts)}"
        )
        self.__groups = groups
        self.config = CFG(deepcopy(_DEFAULT_CONV_CONFIGS))
        self.config.update(deepcopy(config))

        self.add_module(
            "init_convs",
            MultiConv(
                in_channels=self.__in_channels,
                out_channels=init_num_filters,
                filter_lengths=init_filter_lengths,
                subsample_lengths=init_subsample_lengths,
                groups=groups,
                activation=self.config.activation,
            ),
        )

        block_in_channels = init_num_filters[-1]
        for idx, nf in enumerate(self.__num_filters):
            # in the case of ordering of "acb",
            # `out_activation` is indeed `in_activation`
            if idx == 0:
                out_activation = False
            else:
                out_activation = True
            # number of main stream convolution defaults to 2
            if isinstance(nf, int):
                block_out_channels = list(repeat(nf, 2))
            else:
                block_out_channels = list(nf)
            self.add_module(
                f"entry_flow_conv_block_{idx}",
                XceptionMultiConv(
                    in_channels=block_in_channels,
                    num_filters=block_out_channels,
                    filter_lengths=self.__filter_lengths[idx],
                    subsample_length=self.__subsample_lengths[idx],
                    subsample_kernel=self.__subsample_kernels[idx],
                    dilations=self.__dilations[idx],
                    groups=self.__groups,
                    dropouts=self.__dropouts[idx],
                    out_activation=out_activation,
                    **self.config,
                ),
            )
            block_in_channels = block_out_channels[-1]
            if isinstance(self.__block_dropouts[idx], dict):
                if (
                    self.__block_dropouts[idx]["type"] == "1d"
                    and self.__block_dropouts[idx]["p"] > 0
                ):
                    self.add_module(
                        f"entry_flow_dropout_{idx}",
                        nn.Dropout1d(self.__block_dropouts[idx]["p"]),
                    )
                elif (
                    self.__block_dropouts[idx]["type"] is None
                    and self.__block_dropouts[idx]["p"] > 0
                ):
                    self.add_module(
                        f"entry_flow_dropout_{idx}",
                        nn.Dropout(self.__block_dropouts[idx]["p"]),
                    )
            elif self.__block_dropouts[idx] > 0:
                self.add_module(
                    f"entry_flow_dropout_{idx}", nn.Dropout(self.__block_dropouts[idx])
                )

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass of the entry flow.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        Returns
        -------
        output : torch.Tensor
            Output tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        """
        output = super().forward(input)
        return output

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the entry flow.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input tensors.
        batch_size : int, optional
            Batch size of the input tensors.

        Returns
        -------
        output_shape : sequence
            The output shape of the entry flow.

        """
        _seq_len = seq_len
        for module in self:
            if isinstance(module, (nn.Dropout, nn.Dropout1d)):
                continue
            output_shape = module.compute_output_shape(_seq_len, batch_size)
            _, _, _seq_len = output_shape
        return output_shape


class XceptionMiddleFlow(nn.Sequential, SizeMixin):
    """Middle flow of the Xception model.

    Middle flow consists of several Xception blocks of 3 convolutions
    and without sub-sampling.

    Parameters
    ----------
    in_channels : int
        Number of channels in the input tensor.
    num_filters: Sequence[int] or Sequence[Sequence[int]]
        Number of filters (output channels) of the convolutions
        of Xception blocks.
    filter_lengths : int or Sequence[int] or Sequence[Sequence[int]]
        Filter length(s) of the convolutions of Xception blocks.
    dilations : int or Sequence[int] or Sequence[Sequence[int]], default 1
        Dilation(s) of the convolutions of Xception blocks.
    groups : int, default 1
        Connection pattern (of channels) of the inputs and outputs.
    dropouts : float or dict or Sequence[Union[float, dict]] or Sequence[Sequence[Union[float, dict]]], default 0.0
        Dropout(s) after each :class:`Conv_Bn_Activation` blocks
        in the Xception blocks.
    block_dropouts : float or dict or Sequence[Union[float, dict]], default 0.0
        Dropout(s) after the Xception blocks
    config : dict, optional
        Other parameters for Xception blocks, including
        activation choices, weight initializer, batch normalization choices, etc.
        for the convolutional layers,
        and subsampling modes for subsampling layers, etc.

    """

    __name__ = "XceptionMiddleFlow"

    def __init__(
        self,
        in_channels: int,
        num_filters: Union[Sequence[int], Sequence[Sequence[int]]],
        filter_lengths: Union[int, Sequence[int], Sequence[Sequence[int]]],
        dilations: Union[int, Sequence[int], Sequence[Sequence[int]]] = 1,
        groups: int = 1,
        dropouts: Union[
            float,
            dict,
            Sequence[Union[float, dict]],
            Sequence[Sequence[Union[float, dict]]],
        ] = 0.0,
        block_dropouts: Union[float, dict, Sequence[Union[float, dict]]] = 0.0,
        **config,
    ) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.__num_filters = list(num_filters)
        self.__num_blocks = len(self.__num_filters)
        if isinstance(filter_lengths, int):
            self.__filter_lengths = list(repeat(filter_lengths, self.__num_blocks))
        else:
            self.__filter_lengths = list(filter_lengths)
        assert self.__num_blocks == len(self.__filter_lengths), (
            f"the middle flow has {self.__num_blocks} blocks, "
            f"while `filter_lengths` indicates {len(self.__filter_lengths)}"
        )
        if isinstance(dilations, int):
            self.__dilations = list(repeat(dilations, self.__num_blocks))
        else:
            self.__dilations = list(dilations)
        assert self.__num_blocks == len(self.__dilations), (
            f"the middle flow has {self.__num_blocks} blocks, "
            f"while `dilations` indicates {len(self.__dilations)}"
        )
        if isinstance(dropouts, (Real, dict)):
            self.__dropouts = list(repeat(dropouts, self.__num_blocks))
        else:
            self.__dropouts = list(dropouts)
        assert self.__num_blocks == len(self.__dropouts), (
            f"the middle flow has {self.__num_blocks} blocks, "
            f"while `dropouts` indicates {len(self.__dropouts)}"
        )
        if isinstance(block_dropouts, (Real, dict)):
            self.__block_dropouts = list(repeat(block_dropouts, self.__num_blocks))
        else:
            self.__block_dropouts = list(block_dropouts)
        assert self.__num_blocks == len(self.__block_dropouts), (
            f"the middle flow has {self.__num_blocks} blocks, "
            f"while `block_dropouts` indicates {len(self.__block_dropouts)}"
        )
        self.__groups = groups
        self.config = CFG(deepcopy(_DEFAULT_CONV_CONFIGS))
        self.config.update(deepcopy(config))

        block_in_channels = self.__in_channels
        for idx, nf in enumerate(self.__num_filters):
            # number of main stream convolution defaults to 3
            if isinstance(nf, int):
                block_out_channels = list(repeat(nf, 3))
            else:
                block_out_channels = list(nf)
            self.add_module(
                f"middle_flow_conv_block_{idx}",
                XceptionMultiConv(
                    in_channels=block_in_channels,
                    num_filters=block_out_channels,
                    filter_lengths=self.__filter_lengths[idx],
                    dilations=self.__dilations[idx],
                    groups=self.__groups,
                    dropouts=self.__dropouts[idx],
                    **self.config,
                ),
            )
            block_in_channels = block_out_channels[-1]
            if isinstance(self.__block_dropouts[idx], dict):
                if (
                    self.__block_dropouts[idx]["type"] == "1d"
                    and self.__block_dropouts[idx]["p"] > 0
                ):
                    self.add_module(
                        f"middle_flow_dropout_{idx}",
                        nn.Dropout1d(self.__block_dropouts[idx]["p"]),
                    )
                elif (
                    self.__block_dropouts[idx]["type"] is None
                    and self.__block_dropouts[idx]["p"] > 0
                ):
                    self.add_module(
                        f"middle_flow_dropout_{idx}",
                        nn.Dropout(self.__block_dropouts[idx]["p"]),
                    )
            elif self.__block_dropouts[idx] > 0:
                self.add_module(
                    f"middle_flow_dropout_{idx}", nn.Dropout(self.__block_dropouts[idx])
                )

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass of the middle flow.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        Returns
        -------
        output : torch.Tensor
            Output tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        """
        output = super().forward(input)
        return output

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the middle flow.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input tensors.
        batch_size : int, optional
            Batch size of the input tensors.

        Returns
        -------
        output_shape : sequence
            The output shape of the middle flow.

        """
        _seq_len = seq_len
        for module in self:
            if isinstance(module, (nn.Dropout, nn.Dropout1d)):
                continue
            output_shape = module.compute_output_shape(_seq_len, batch_size)
            _, _, _seq_len = output_shape
        return output_shape


class XceptionExitFlow(nn.Sequential, SizeMixin):
    """Exit flow of the Xception model.

    Exit flow consists of several Xception blocks of 2 convolutions,
    followed by several separable convolutions.

    Parameters
    ----------
    in_channels : int
        Number of channels in the input tensor.
    final_num_filters : Sequence[int]
        Number of filters (output channels) of the final convolutions.
    final_filter_lengths : int or Sequence[int]
        Filter length(s) of the convolutions of the final convolutions.
    final_subsample_lengths : int or Sequence[int]
        Subsampling length(s) (stride(s)) of the final convolutions.
    num_filters : Sequence[int] or Sequence[Sequence[int]]
        Number of filters of the convolutions of Xception blocks.
    filter_lengths : int or Sequence[int] or Sequence[Sequence[int]]
        Filter length(s) of the convolutions of Xception blocks.
    subsample_lengths : int or Sequence[int]
        Subsampling length(s) of the Xception blocks.
    subsample_kernels : int or Sequence[int], optional
        Subsampling kernel size(s) of the Xception blocks.
    dilations: int or Sequence[int] or Sequence[Sequence[int]], default 1
        Dilation(s) of the convolutions of Xception blocks.
    groups : int, default 1
        Connection pattern (of channels) of the inputs and outputs.
    dropouts : float or dict or Sequence[Union[float, dict]] or Sequence[Sequence[Union[float, dict]]], default 0.0
        Dropout(s) after each :class:`Conv_Bn_Activation` blocks
        in the Xception blocks.
    block_dropouts : float or dict or Sequence[Union[float, dict]], default 0.0
        Dropout(s) after each of the Xception blocks
        and each of the final convolutions.
    config : dict, optional
        Other parameters for Xception blocks and final convolutions, including
        activation choices, weight initializer, batch normalization choices, etc.
        for the convolutional layers,
        and subsampling modes for subsampling layers, etc.

    """

    __name__ = "XceptionExitFlow"

    def __init__(
        self,
        in_channels: int,
        final_num_filters: Sequence[int],
        final_filter_lengths: Union[int, Sequence[int]],
        num_filters: Union[Sequence[int], Sequence[Sequence[int]]],
        filter_lengths: Union[int, Sequence[int], Sequence[Sequence[int]]],
        subsample_lengths: Union[int, Sequence[int]],
        subsample_kernels: Optional[Union[int, Sequence[int]]] = None,
        dilations: Union[int, Sequence[int], Sequence[Sequence[int]]] = 1,
        groups: int = 1,
        dropouts: Union[
            float,
            dict,
            Sequence[Union[float, dict]],
            Sequence[Sequence[Union[float, dict]]],
        ] = 0.0,
        block_dropouts: Union[float, dict, Sequence[Union[float, dict]]] = 0.0,
        **config,
    ) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.__num_filters = list(num_filters)
        self.__num_blocks = len(self.__num_filters)
        if isinstance(filter_lengths, int):
            self.__filter_lengths = list(repeat(filter_lengths, self.__num_blocks))
        else:
            self.__filter_lengths = list(filter_lengths)
        assert self.__num_blocks == len(
            self.__filter_lengths
        ), f"the exit flow has {self.__num_blocks} blocks, while `filter_lengths` indicates {len(self.__filter_lengths)}"
        if isinstance(subsample_lengths, int):
            self.__subsample_lengths = list(
                repeat(subsample_lengths, self.__num_blocks)
            )
        else:
            self.__subsample_lengths = list(subsample_lengths)
        assert self.__num_blocks == len(self.__subsample_lengths), (
            f"the exit flow has {self.__num_blocks} blocks, "
            f"while `subsample_lengths` indicates {len(self.__subsample_lengths)}"
        )
        if subsample_kernels is None:
            self.__subsample_kernels = deepcopy(self.__subsample_lengths)
        elif isinstance(subsample_kernels, int):
            self.__subsample_kernels = list(
                repeat(subsample_kernels, self.__num_blocks)
            )
        else:
            self.__subsample_kernels = list(subsample_kernels)
        assert self.__num_blocks == len(self.__subsample_kernels), (
            f"the exit flow has {self.__num_blocks} blocks, "
            f"while `subsample_kernels` indicates {len(self.__subsample_kernels)}"
        )
        if isinstance(dilations, int):
            self.__dilations = list(repeat(dilations, self.__num_blocks))
        else:
            self.__dilations = list(dilations)
        assert self.__num_blocks == len(self.__dilations), (
            f"the exit flow has {self.__num_blocks} blocks, "
            f"while `dilations` indicates {len(self.__dilations)}"
        )
        if isinstance(dropouts, (Real, dict)):
            self.__dropouts = list(repeat(dropouts, self.__num_blocks))
        else:
            self.__dropouts = list(dropouts)
        assert self.__num_blocks == len(self.__dropouts), (
            f"the exit flow has {self.__num_blocks} blocks, "
            f"while `dropouts` indicates {len(self.__dropouts)}"
        )
        if isinstance(block_dropouts, (Real, dict)):
            self.__block_dropouts = list(
                repeat(block_dropouts, self.__num_blocks + len(final_num_filters))
            )
        else:
            self.__block_dropouts = list(block_dropouts)
        assert self.__num_blocks + len(final_num_filters) == len(
            self.__block_dropouts
        ), (
            f"the exit flow has {self.__num_blocks + len(final_num_filters)} blocks, "
            f"including the final convolutions, while `block_dropouts` indicates {len(self.__block_dropouts)}"
        )
        self.__groups = groups
        self.config = CFG(deepcopy(_DEFAULT_CONV_CONFIGS))
        self.config.update(deepcopy(config))

        block_in_channels = self.__in_channels
        for idx, nf in enumerate(self.__num_filters):
            # number of main stream convolution defaults to 2
            if isinstance(nf, int):
                block_out_channels = list(repeat(nf, 2))
            else:
                block_out_channels = list(nf)
            self.add_module(
                f"exit_flow_conv_block_{idx}",
                XceptionMultiConv(
                    in_channels=block_in_channels,
                    num_filters=block_out_channels,
                    filter_lengths=self.__filter_lengths[idx],
                    subsample_length=self.__subsample_lengths[idx],
                    subsample_kernel=self.__subsample_kernels[idx],
                    dilations=self.__dilations[idx],
                    groups=self.__groups,
                    dropouts=self.__dropouts[idx],
                    **self.config,
                ),
            )
            block_in_channels = block_out_channels[-1]
            if isinstance(self.__block_dropouts[idx], dict):
                if (
                    self.__block_dropouts[idx]["type"] == "1d"
                    and self.__block_dropouts[idx]["p"] > 0
                ):
                    self.add_module(
                        f"exit_flow_dropout_{idx}",
                        nn.Dropout1d(self.__block_dropouts[idx]["p"]),
                    )
                elif (
                    self.__block_dropouts[idx]["type"] is None
                    and self.__block_dropouts[idx]["p"] > 0
                ):
                    self.add_module(
                        f"exit_flow_dropout_{idx}",
                        nn.Dropout(self.__block_dropouts[idx]["p"]),
                    )
            elif self.__block_dropouts[idx] > 0:
                self.add_module(
                    f"exit_flow_dropout_{idx}", nn.Dropout(self.__block_dropouts[idx])
                )

        self.add_module(
            "final_convs",
            MultiConv(
                in_channels=block_in_channels,
                out_channels=final_num_filters,
                filter_lengths=final_filter_lengths,
                groups=groups,
                conv_type="separable",
                activation=self.config.activation,
            ),
        )

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass of the exit flow.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        Returns
        -------
        output : torch.Tensor
            Output tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        """
        output = super().forward(input)
        return output

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the exit flow.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input tensors.
        batch_size : int, optional
            Batch size of the input tensors.

        Returns
        -------
        output_shape : sequence
            The output shape of the exit flow.

        """
        _seq_len = seq_len
        for module in self:
            if isinstance(module, (nn.Dropout, nn.Dropout1d)):
                continue
            output_shape = module.compute_output_shape(_seq_len, batch_size)
            _, _, _seq_len = output_shape
        return output_shape


class Xception(nn.Sequential, SizeMixin, CitationMixin):
    """Xception model.

    Xception is an architecture that uses depthwise separable convolutions
    to build light-weight deep neural networks, as described in [1]_.
    Its official implementation is available in [2]_, and a PyTorch
    implementation is available in [3]_. Xception is currently not widely
    used in the field of ECG analysis, but has the potential to be highly
    effective for this task.

    Parameters
    ----------
    in_channels : int
        Number of channels in the input.
    config : dict
        Other hyper-parameters of the Module, ref. corr. config file.
        For keyword arguments that must be set in 3 sub-dict,
        namely in "entry_flow", "middle_flow", and "exit_flow",
        refer to corr. docstring of each class.

    References
    ----------
    .. [1] Chollet, François. "Xception: Deep learning with depthwise separable convolutions."
           Proceedings of the IEEE conference on computer vision and pattern recognition. 2017.
    .. [2] https://github.com/keras-team/keras-applications/blob/master/keras_applications/xception.py
    .. [3] https://github.com/Cadene/pretrained-models.pytorch/blob/master/pretrainedmodels/models/xception.py

    """

    __name__ = "Xception"

    def __init__(self, in_channels: int, **config) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.config = CFG(deepcopy(config))

        entry_flow_in_channels = self.__in_channels
        entry_flow = XceptionEntryFlow(
            in_channels=entry_flow_in_channels, **(self.config.entry_flow)
        )
        self.add_module("entry_flow", entry_flow)

        _, middle_flow_in_channels, _ = entry_flow.compute_output_shape()
        middle_flow = XceptionMiddleFlow(
            in_channels=middle_flow_in_channels, **(self.config.middle_flow)
        )
        self.add_module("middle_flow", middle_flow)

        _, exit_flow_in_channels, _ = middle_flow.compute_output_shape()
        exit_flow = XceptionExitFlow(
            in_channels=exit_flow_in_channels, **(self.config.exit_flow)
        )
        self.add_module(
            "exit_flow",
            exit_flow,
        )

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass of the model.

        Parameters
        ----------
        input : torch.Tensor.
            Input signal tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        Returns
        -------
        output : torch.Tensor
            Output tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        """
        output = super().forward(input)
        return output

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape the model."""
        return compute_sequential_output_shape(self, seq_len, batch_size)

    @property
    def in_channels(self) -> int:
        return self.__in_channels

    @property
    def doi(self) -> List[str]:
        return list(set(self.config.get("doi", []) + ["10.1109/cvpr.2017.195"]))



"""
2nd place (entry 0433) of CPSC2019

the main differences to a normal Unet are that

1. at the bottom, subtraction (and concatenation) is used
2. uses triple convolutions at each block, instead of double convolutions
3. dropout is used between certain convolutional layers ("cba" layers indeed)

"""

import textwrap
import warnings
from copy import deepcopy
from itertools import repeat
from typing import Optional, Union, Sequence, List

import numpy as np
import torch
from torch import Tensor, nn

from ...cfg import CFG
from ...models._nets import (
    BranchedConv,
    Conv_Bn_Activation,
    DownSample,
    MultiConv,
)
from ...utils.misc import add_docstring
from ...utils.utils_nn import (
    CkptMixin,
    SizeMixin,
    compute_deconv_output_shape,
    compute_sequential_output_shape,
    compute_sequential_output_shape_docstring,
)
from ...model_configs import ECG_SUBTRACT_UNET_CONFIG


__all__ = [
    "ECG_SUBTRACT_UNET",
]


class TripleConv(MultiConv):
    """Triple convolutional layer.

    CBA --> (Dropout) --> CBA --> (Dropout) --> CBA --> (Dropout).

    Parameters
    ----------
    in_channels : int
        Number of channels in the input tensor.
    out_channels : int or Sequence[int]
        Number of channels produced by the (last) convolutional layer(s).
    filter_lengths : int or Sequence[int]
        Length(s) of the filters (kernel size).
    subsample_lengths : int or Sequence[int], default 1
        Subsample length(s) (stride(s)) of the convolutions
    groups: int, default 1,
        connection pattern (of channels) of the inputs and outputs.
    dropouts : float or dict or Sequence[Union[float, dict]], default 0.0
        Dropout ratio after each :class:`Conv_Bn_Activation` block.
    out_activation : bool, default True
        If True, the last mini-block of :class:`Conv_Bn_Activation`
        will have activation as in `config`; otherwise, no activation.
    config : dict
        Other hyper-parameters, including
        activation choices, weight initializer, batch normalization choices, etc.
        for the convolutional layers.

    """

    __name__ = "TripleConv"

    def __init__(
        self,
        in_channels: int,
        out_channels: Union[Sequence[int], int],
        filter_lengths: Union[Sequence[int], int],
        subsample_lengths: Union[Sequence[int], int] = 1,
        groups: int = 1,
        dropouts: Union[Sequence[Union[float, dict]], float, dict] = 0.0,
        out_activation: bool = True,
        **config,
    ) -> None:
        _num_convs = 3
        if isinstance(out_channels, int):
            _out_channels = list(repeat(out_channels, _num_convs))
        else:
            _out_channels = list(out_channels)
            assert _num_convs == len(_out_channels)

        super().__init__(
            in_channels=in_channels,
            out_channels=_out_channels,
            filter_lengths=filter_lengths,
            subsample_lengths=subsample_lengths,
            groups=groups,
            dropouts=dropouts,
            out_activation=out_activation,
            **config,
        )


class DownTripleConv(nn.Sequential, SizeMixin):
    """Down sampling block of the U-net architecture.

    Composed of a down sampling layer and 3 convolutional layers.

    Parameters
    ----------
    down_scale : int
        Down sampling scale.
    in_channels : int
        Number of channels in the input.
    out_channels : int or Sequence[int]
        Number of channels produced by the (last) convolutional layer(s).
    filter_lengths : int or Sequence[int]
        Length(s) of the filters (kernel size).
    groups : int, default 1
        Connection pattern (of channels) of the inputs and outputs.
    dropouts : float or dict or Sequence[Union[float, dict]], default 0.0
        Dropout ratio after each :class:`Conv_Bn_Activation` block.
    mode : str, default "max"
        Down sampling mode, one of {:class:`DownSample`.__MODES__}.
    config : dict
        Other parameters, including
        activation choices, weight initializer, batch normalization choices, etc.
        for the convolutional layers.

    """

    __name__ = "DownTripleConv"
    __MODES__ = deepcopy(DownSample.__MODES__)

    def __init__(
        self,
        down_scale: int,
        in_channels: int,
        out_channels: Union[Sequence[int], int],
        filter_lengths: Union[Sequence[int], int],
        groups: int = 1,
        dropouts: Union[Sequence[Union[float, dict]], float, dict] = 0.0,
        mode: str = "max",
        **config,
    ) -> None:
        super().__init__()
        self.__mode = mode.lower()
        assert self.__mode in self.__MODES__
        self.__down_scale = down_scale
        self.__in_channels = in_channels
        self.__out_channels = out_channels
        self.config = CFG(deepcopy(config))

        self.add_module(
            "down_sample",
            DownSample(
                down_scale=self.__down_scale,
                in_channels=self.__in_channels,
                norm=False,
                mode=mode,
            ),
        )
        self.add_module(
            "triple_conv",
            TripleConv(
                in_channels=self.__in_channels,
                out_channels=self.__out_channels,
                filter_lengths=filter_lengths,
                subsample_lengths=1,
                groups=groups,
                dropouts=dropouts,
                **(self.config),
            ),
        )

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, channels, seq_len)``.

        Returns
        -------
        out : torch.Tensor
            Output tensor,
            of shape ``(batch_size, channels, seq_len)``.

        """
        out = super().forward(input)
        return out

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the down sampling block."""
        return compute_sequential_output_shape(self, seq_len, batch_size)


class DownBranchedDoubleConv(nn.Module, SizeMixin):
    """The bottom block of the encoder in the U-Net architecture.

    Parameters
    ----------
    down_scale : int
        Down sampling scale.
    in_channels : int
        Number of channels in the input tensor.
    out_channels : Sequence[Sequence[int]]
        Number of channels produced by the (last) convolutional layer(s).
    filter_lengths : int or Sequence[int] or Sequence[Sequence[int]]
        Length(s) of the filters (kernel size).
    dilations : int or Sequence[int] or Sequence[Sequence[int]], default 1
        Dilation(s) of the convolutions.
    groups : int, default 1
        Connection pattern (of channels) of the inputs and outputs.
    dropouts : float or dict or Sequence[Union[float, dict]], default 0.0
        Dropout ratio after each :class:`Conv_Bn_Activation` block.
    mode : str, default "max"
        Down sampling mode, one of {:class:`DownSample`.__MODES__}.
    config: dict
        Other hyper-parameters, including
        activation choices, weight initializer, batch normalization choices, etc.
        for the convolutional layers.

    """

    __name__ = "DownBranchedDoubleConv"
    __MODES__ = deepcopy(DownSample.__MODES__)

    def __init__(
        self,
        down_scale: int,
        in_channels: int,
        out_channels: Sequence[Sequence[int]],
        filter_lengths: Union[Sequence[Sequence[int]], Sequence[int], int],
        dilations: Union[Sequence[Sequence[int]], Sequence[int], int] = 1,
        groups: int = 1,
        dropouts: Union[Sequence[Union[float, dict]], float, dict] = 0.0,
        mode: str = "max",
        **config,
    ) -> None:
        super().__init__()
        self.__mode = mode.lower()
        assert self.__mode in self.__MODES__
        self.__down_scale = down_scale
        self.__in_channels = in_channels
        self.__out_channels = out_channels
        self.config = CFG(deepcopy(config))

        self.down_sample = DownSample(
            down_scale=self.__down_scale,
            in_channels=self.__in_channels,
            norm=False,
            mode=mode,
        )
        self.branched_conv = BranchedConv(
            in_channels=self.__in_channels,
            out_channels=self.__out_channels,
            filter_lengths=filter_lengths,
            subsample_lengths=1,
            dilations=dilations,
            groups=groups,
            dropouts=dropouts,
            **(self.config),
        )

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, channels, seq_len)``.

        Returns
        -------
        out : torch.Tensor
            Output tensor,
            of shape ``(batch_size, channels, seq_len)``.

        """
        out = self.down_sample(input)
        out = self.branched_conv(out)
        # SUBTRACT
        # currently (micro scope) - (macro scope)
        # TODO: consider (macro scope) - (micro scope)
        out.append(out[0] - out[1])
        out = torch.cat(out, dim=1)  # concate along the channel axis
        return out

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the block.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input tensor.
        batch_size : int, optional
            Batch size of the input tensor.

        Returns
        -------
        output_shape : sequence
            Output shape of the block.

        """
        _seq_len = seq_len
        output_shape = self.down_sample.compute_output_shape(seq_len=_seq_len)
        _, _, _seq_len = output_shape
        output_shapes = self.branched_conv.compute_output_shape(seq_len=_seq_len)
        # output_shape = output_shapes[0][0], sum([s[1] for s in output_shapes]), output_shapes[0][-1]
        n_branches = len(output_shapes)
        output_shape = (
            output_shapes[0][0],
            (n_branches + 1) * output_shapes[0][1],
            output_shapes[0][-1],
        )
        return output_shape


class UpTripleConv(nn.Module, SizeMixin):
    """Up sampling block in the U-Net architecture.

    Upscaling then double conv, with input of corr. down layer concatenated
    up sampling --> conv (conv --> (dropout -->) conv --> (dropout -->) conv)
        ^
        |
    extra input

    Channels are shrinked after up sampling.

    Parameters
    ----------
    up_scale : int
        Scale of up sampling.
    in_channels : int
        Number of channels in the input tensor.
    out_channels : int
        Number of channels produced by the convolutional layers.
    filter_lengths : int or Sequence[int]
        Length(s) of the filters (kernel size) of the convolutional layers.
    deconv_filter_length : int
        Length(s) of the filters (kernel size) of the
        deconvolutional upsampling layer,
        used when `mode` is ``"deconv"``.
    groups : int, default 1
        Connection pattern (of channels) of the inputs and outputs.
        Not used currently.
    dropouts : float or dict or Sequence[Union[float, dict]], default 0.0
        Dropout ratio after each :class:`Conv_Bn_Activation` block.
    mode : str, default "deconv"
        Mode of up sampling, case insensitive. Should be ``"deconv"``
        or methods supported by :class:`torch.nn.Upsample`.
    config : dict
        Other hyper-parameters, including
        activation choices, weight initializer, batch normalization choices, etc.
        for the deconvolutional layers.

    """

    __name__ = "UpTripleConv"
    __MODES__ = [
        "nearest",
        "linear",
        "area",
        "deconv",
    ]

    def __init__(
        self,
        up_scale: int,
        in_channels: int,
        out_channels: int,
        filter_lengths: Union[Sequence[int], int],
        deconv_filter_length: Optional[int] = None,
        groups: int = 1,
        dropouts: Union[Sequence[Union[float, dict]], float, dict] = 0.0,
        mode: str = "deconv",
        **config,
    ) -> None:
        super().__init__()
        self.__up_scale = up_scale
        self.__in_channels = in_channels
        self.__out_channels = out_channels
        self.__deconv_filter_length = deconv_filter_length
        self.__mode = mode.lower()
        assert self.__mode in self.__MODES__
        self.config = CFG(deepcopy(config))

        # the following has to be checked
        # if bilinear, use the normal convolutions to reduce the number of channels
        if self.__mode == "deconv":
            self.__deconv_padding = max(
                0, (self.__deconv_filter_length - self.__up_scale) // 2
            )
            self.up = nn.ConvTranspose1d(
                in_channels=self.__in_channels,
                out_channels=self.__in_channels,
                kernel_size=self.__deconv_filter_length,
                stride=self.__up_scale,
                padding=self.__deconv_padding,
            )
        else:
            self.up = nn.Upsample(
                scale_factor=self.__up_scale,
                mode=mode,
            )
        self.conv = TripleConv(
            # `+ self.__out_channels` corr. to the output of the corr. down layer
            in_channels=self.__in_channels + self.__out_channels[-1],
            out_channels=self.__out_channels,
            filter_lengths=filter_lengths,
            subsample_lengths=1,
            groups=groups,
            dropouts=dropouts,
            **(self.config),
        )

    def forward(self, input: Tensor, down_output: Tensor) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor from the previous up sampling block.
        down_output : torch.Tensor
            Input tensor of the last layer of corr. down sampling block.

        Returns
        -------
        output : torch.Tensor
            Output tensor of this block.

        """
        output = self.up(input)
        output = torch.cat(
            [down_output, output], dim=1
        )  # concate along the channel axis
        output = self.conv(output)

        return output

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the block.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input tensor.
        batch_size : int, optional
            Batch size of the input tensor.

        Returns
        -------
        output_shape : sequence
            Output shape of the block.

        """
        _sep_len = seq_len
        if self.__mode == "deconv":
            output_shape = compute_deconv_output_shape(
                input_shape=[batch_size, self.__in_channels, _sep_len],
                num_filters=self.__in_channels,
                kernel_size=self.__deconv_filter_length,
                stride=self.__up_scale,
                padding=self.__deconv_padding,
            )
        else:
            output_shape = [batch_size, self.__in_channels, self.__up_scale * _sep_len]
        _, _, _seq_len = output_shape
        output_shape = self.conv.compute_output_shape(_seq_len, batch_size)
        return output_shape


class ECG_SUBTRACT_UNET(nn.Module, CkptMixin, SizeMixin):
    """U-Net for ECG wave delineation.

    Entry 0433 of CPSC2019, which is a modification of the U-Net
    using subtraction instead of addition in branched bottom block.

    Parameters
    ----------
    classes : Sequence[str]
        List of names of the classes.
    n_leads : int
        Number of input leads (number of input channels).
    config : CFG, optional
        Other hyper-parameters, including kernel sizes, etc.
        Refer to the corresponding config file.

    """

    __name__ = "ECG_SUBTRACT_UNET"

    def __init__(
        self,
        classes: Sequence[str],
        n_leads: int,
        config: Optional[CFG] = None,
    ) -> None:
        super().__init__()
        self.classes = list(classes)
        self.n_classes = len(classes)
        self.__out_channels = len(classes)
        self.__in_channels = n_leads
        self.config = deepcopy(ECG_SUBTRACT_UNET_CONFIG)
        if not config:
            warnings.warn(
                "No config is provided, using default config.", RuntimeWarning
            )
        self.config.update(deepcopy(config) or {})

        # TODO: an init batch normalization?
        if self.config.init_batch_norm:
            self.init_bn = nn.BatchNorm1d(
                num_features=self.__in_channels,
                eps=1e-5,  # default val
                momentum=0.1,  # default val
            )

        self.init_conv = TripleConv(
            in_channels=self.__in_channels,
            out_channels=self.config.init_num_filters,
            filter_lengths=self.config.init_filter_length,
            subsample_lengths=1,
            groups=self.config.groups,
            dropouts=self.config.init_dropouts,
            batch_norm=self.config.batch_norm,
            activation=self.config.activation,
            kw_activation=self.config.kw_activation,
            kernel_initializer=self.config.kernel_initializer,
            kw_initializer=self.config.kw_initializer,
        )

        self.down_blocks = nn.ModuleDict()
        in_channels = self.config.init_num_filters
        for idx in range(self.config.down_up_block_num - 1):
            self.down_blocks[f"down_{idx}"] = DownTripleConv(
                down_scale=self.config.down_scales[idx],
                in_channels=in_channels,
                out_channels=self.config.down_num_filters[idx],
                filter_lengths=self.config.down_filter_lengths[idx],
                groups=self.config.groups,
                dropouts=self.config.down_dropouts[idx],
                mode=self.config.down_mode,
                **(self.config.down_block),
            )
            in_channels = self.config.down_num_filters[idx][-1]

        self.bottom_block = DownBranchedDoubleConv(
            down_scale=self.config.down_scales[-1],
            in_channels=in_channels,
            out_channels=self.config.bottom_num_filters,
            filter_lengths=self.config.bottom_filter_lengths,
            dilations=self.config.bottom_dilations,
            groups=self.config.groups,
            dropouts=self.config.bottom_dropouts,
            mode=self.config.down_mode,
            **(self.config.down_block),
        )

        self.up_blocks = nn.ModuleDict()
        # in_channels = sum([branch[-1] for branch in self.config.bottom_num_filters])
        in_channels = self.bottom_block.compute_output_shape(None, None)[1]
        for idx in range(self.config.down_up_block_num):
            self.up_blocks[f"up_{idx}"] = UpTripleConv(
                up_scale=self.config.up_scales[idx],
                in_channels=in_channels,
                out_channels=self.config.up_num_filters[idx],
                filter_lengths=self.config.up_conv_filter_lengths[idx],
                deconv_filter_length=self.config.up_deconv_filter_lengths[idx],
                groups=self.config.groups,
                mode=self.config.up_mode,
                dropouts=self.config.up_dropouts[idx],
                **(self.config.up_block),
            )
            in_channels = self.config.up_num_filters[idx][-1]

        self.out_conv = Conv_Bn_Activation(
            in_channels=self.config.up_num_filters[-1][-1],
            out_channels=self.__out_channels,
            kernel_size=self.config.out_filter_length,
            stride=1,
            groups=self.config.groups,
            norm=self.config.get("out_norm", self.config.get("out_batch_norm")),
            activation=None,
            kernel_initializer=self.config.kernel_initializer,
            kw_initializer=self.config.kw_initializer,
        )

        # for inference
        # if background counted in `classes`, use softmax
        # otherwise use sigmoid
        self.softmax = nn.Softmax(-1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass of the model.

        Parameters
        ----------
        input : torch.Tensor
            Input signal tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        Returns
        -------
        output : torch.Tensor
            Output tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        """
        if self.config.init_batch_norm:
            x = self.init_bn(input)
        else:
            x = input

        # down
        to_concat = [self.init_conv(x)]
        for idx in range(self.config.down_up_block_num - 1):
            to_concat.append(self.down_blocks[f"down_{idx}"](to_concat[-1]))
        to_concat.append(self.bottom_block(to_concat[-1]))

        # up
        up_input = to_concat[-1]
        to_concat = to_concat[-2::-1]
        for idx in range(self.config.down_up_block_num):
            up_output = self.up_blocks[f"up_{idx}"](up_input, to_concat[idx])
            up_input = up_output

        # output
        output = self.out_conv(up_output)

        # to keep in accordance with other models
        # (batch_size, channels, seq_len) --> (batch_size, seq_len, channels)
        output = output.permute(0, 2, 1)

        # TODO: consider adding CRF at the tail to make final prediction

        return output

    @torch.no_grad()
    def inference(
        self, input: Union[np.ndarray, Tensor], bin_pred_thr: float = 0.5
    ) -> Tensor:
        """Method for making inference on a single input."""
        raise NotImplementedError("implement a task specific inference method")

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the model.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input signal tensor.
        batch_size : int, optional
            Batch size of the input signal tensor.

        Returns
        -------
        output_shape : sequence
            Output shape of the model.

        """
        output_shape = (batch_size, seq_len, self.n_classes)
        return output_shape

    @property
    def doi(self) -> List[str]:
        # TODO: add doi
        return list(set(self.config.get("doi", []) + []))



"""
Computes the respiratory rate from single-lead ECG signals.

A python re-implementation of the `edr` function of physionet edr.c
"""

from numbers import Real
from typing import Sequence

import numpy as np


__all__ = [
    "phs_edr",
]


def phs_edr(
    sig: Sequence,
    fs: int,
    rpeaks: Sequence,
    winL_t: Real = 40,
    winR_t: Real = 40,
    return_with_time: bool = True,
    mode: str = "complex",
    verbose: int = 0,
) -> np.ndarray:
    """
    computes the respiratory rate from single-lead ECG signals.

    ref. the `main` function and the `getxy`, `edr` function of physionet edr.c

    Parameters
    ----------
    sig: array-like,
        the (single lead ECG) signal
    fs: int,
        sampling frequency of the signal
    rpeaks: array-like,
        indices of R peaks in the signal
    winL_t: numbers.Real, default 40,
        left length of the window at R peaks for the computation of the area of a QRS complex,
        with units in milliseconds
    winR_t: numbers.Real, default 40,
        right length of the window at R peaks for the computation of the area of a QRS complex,
        with units in milliseconds
    return_with_time: bool, default True,
        if True, returns the time along with the EDR values at which they are computed
    mode: str, default "complex", can also be "simple",
        apply a filtering process (the `edr` function of physionet edr.c)
        or simply use the `_getxy` function to compute EDR
    verbose: int, default 0,
        for printing the computation details

    Returns
    -------
    np.ndarray,
        1d, if `return_with_time` is set False,
        2d in the form of [ts, val] (ts in milliseconds), if `return_with_time` is set True

    """
    ts = np.array(rpeaks) * 1000 // fs
    winL, winR = int(winL_t * fs / 1000), int(winR_t * fs / 1000)

    if mode == "simple":
        if len(rpeaks) > 0:
            ecg_der_rsp = np.vectorize(lambda idx: _getxy(sig, idx - winL, idx + winR))(
                np.array(rpeaks)
            )
        else:
            ecg_der_rsp = np.array([])
    elif mode == "complex":
        ecg_der_rsp = []
        xm, xc, xd, xdmax = 0, 0, 0, 0
        for idx in rpeaks:
            if verbose == -1:
                print("-" * 80)
                print(f"idx = {idx}, winL = {winL}, winR = {winR}")
            x = _getxy(sig, idx - winL, idx + winR)
            if verbose == -1:
                print(f"x = {x}")

            # calculate instantaneous EDR
            if x == 0:
                ecg_der_rsp.append(0)
                continue

            d = x - xm
            if verbose == -1:
                print(f"before: d = {d}, xc = {xc}, xdmax = {xdmax}")

            if xc < 500:  # why 500?
                xc += 1
                dn = d / xc
            else:
                dn = d / xc
                if dn > xdmax:
                    dn = xdmax
                elif dn < -xdmax:
                    dn = -xdmax
            if verbose == -1:
                print(f"after: d = {d}, xc = {xc}, dn = {dn}, xdmax = {xdmax}")

            xm += dn
            xd += abs(dn) - xd / xc

            if xd < 1.0:
                xd = 1.0
            xdmax = 3.0 * xd / xc
            r = d / xd
            if verbose == -1:
                print(f"xm = {xm}, xc = {xc}, xd = {xd}, xdmax = {xdmax}, r = {r}")
            ecg_der_rsp.append(int(r * 50))
            # end of calculation of instantaneous EDR
    else:
        raise ValueError(f"No mode named `{mode}`!")

    ecg_der_rsp = np.array(ecg_der_rsp)

    if verbose >= 2:
        pass  # TODO: some plot

    if return_with_time:
        return np.column_stack((ts, ecg_der_rsp))
    else:
        return ecg_der_rsp


def _getxy(sig: Sequence, von: int, bis: int) -> Real:
    """compute the integrand from `von` to `bis` of the signals with baseline removed"""
    return (np.array(sig)[von : bis + 1]).sum()



"""
"""

from numbers import Real
from random import randint
from typing import Any, List, Optional, Sequence, Tuple, Union

import numpy as np
import torch
from torch import Tensor

from .base import Augmenter

__all__ = [
    "RandomMasking",
]


class RandomMasking(Augmenter):
    """Randomly mask ECGs with a probability.

    Parameters
    ----------
    fs : int
        Sampling frequency of the ECGs to be augmented.
    mask_value : numbers.Real, default 0.0
        Value to mask with.
    mask_width : Sequence[numbers.Real], default ``[0.08, 0.18]``
        Width range of the masking window, with units in seconds
    prob : numbers.Real or Sequence[numbers.Real], default ``[0.3, 0.15]``
        Probabilities of masking ECG signals,
        the first probality is for the batch dimension,
        the second probability is for the lead dimension.
        Note that 0.15 is approximately the proportion of QRS complexes in ECGs.
    inplace : bool, default True
        Whether to mask inplace or not.
    kwargs : dict, optional
        Additional keyword arguments.

    Examples
    --------
    .. code-block:: python

        rm = RandomMasking(fs=500, prob=0.7)
        sig = torch.randn(32, 12, 5000)
        critical_points = [np.arange(250, 5000 - 250, step=400) for _ in range(32)]
        sig, _ = rm(sig, None, critical_points=critical_points)

    """

    __name__ = "RandomMasking"

    def __init__(
        self,
        fs: int,
        mask_value: Real = 0.0,
        mask_width: Sequence[Real] = [0.08, 0.18],
        prob: Union[Sequence[Real], Real] = [0.3, 0.15],
        inplace: bool = True,
        **kwargs: Any
    ) -> None:
        super().__init__()
        self.fs = fs
        self.prob = prob
        if isinstance(self.prob, Real):
            self.prob = np.array([self.prob, self.prob])
        else:
            self.prob = np.array(self.prob)
        assert (self.prob >= 0).all() and (
            self.prob <= 1
        ).all(), "Probability must be between 0 and 1"
        self.mask_value = mask_value
        self.mask_width = (np.array(mask_width) * self.fs).round().astype(int)
        self.inplace = inplace

    def forward(
        self,
        sig: Tensor,
        label: Optional[Tensor],
        *extra_tensors: Sequence[Tensor],
        critical_points: Optional[Sequence[Sequence[int]]] = None,
        **kwargs: Any
    ) -> Tuple[Tensor, ...]:
        """Forward method of the RandomMasking augmenter.

        Parameters
        ----------
        sig : torch.Tensor
            Batched ECGs to be augmented, of shape ``(batch, lead, siglen)``.
        label : torch.Tensor
            Label tensor of the ECGs.
            Not used, but kept for compatibility with other augmenters.
        extra_tensors : Sequence[torch.Tensor], optional
            Not used, but kept for consistency with other augmenters.
        critical_points : Sequence[Sequence[int]], optional
            If given, random masking will be performed in windows centered at these points.
            This is useful for example when one wants to randomly mask QRS complexes.
        kwargs : dict, optional
            Not used, but kept for consistency with other augmenters.

        Returns
        -------
        sig : torch.Tensor
            The augmented ECGs, of shape ``(batch, lead, siglen)``.
        label : torch.Tensor
            Label tensor of the augmented ECGs, unchanged.
        extra_tensors : Sequence[torch.Tensor], optional
            Unchanged extra tensors.

        """
        batch, lead, siglen = sig.shape
        if not self.inplace:
            sig = sig.clone()
        if self.prob[0] == 0:
            return (sig, label, *extra_tensors)
        sig_mask_prob = self.prob[1] / self.mask_width[1]
        sig_mask_scale_ratio = min(self.prob[1] / 4, 0.1) / self.mask_width[1]
        mask = torch.full_like(sig, 1, dtype=sig.dtype, device=sig.device)
        for batch_idx in self.get_indices(prob=self.prob[0], pop_size=batch):
            if critical_points is not None:
                indices = self.get_indices(
                    prob=self.prob[1], pop_size=len(critical_points[batch_idx])
                )
                indices = np.arange(siglen)[indices]
            else:
                indices = np.array(
                    self.get_indices(
                        prob=sig_mask_prob,
                        pop_size=siglen - self.mask_width[1],
                        scale_ratio=sig_mask_scale_ratio,
                    )
                )
                indices += self.mask_width[1] // 2
            for j in indices:
                masked_radius = randint(self.mask_width[0], self.mask_width[1]) // 2
                mask[
                    batch_idx, :, j - masked_radius : j + masked_radius
                ] = self.mask_value
            # print(f"batch_idx = {batch_idx}, len(indices) = {len(indices)}")
        sig = sig.mul_(mask)
        return (sig, label, *extra_tensors)

    def extra_repr_keys(self) -> List[str]:
        return [
            "fs",
            "mask_value",
            "mask_width",
            "prob",
            "inplace",
        ] + super().extra_repr_keys()



# -*- coding: utf-8 -*-

import json
from numbers import Real
from pathlib import Path
from typing import Any, Optional, Sequence, Union, List

import numpy as np
import pandas as pd
import scipy.signal as SS
from scipy.io import loadmat

from ...cfg import DEFAULTS
from ...utils.misc import add_docstring
from ..base import DEFAULT_FIG_SIZE_PER_SEC, CPSCDataBase, DataBaseInfo


__all__ = [
    "CPSC2019",
    "compute_metrics",
]


_CPSC2019_INFO = DataBaseInfo(
    title="""
    The 2nd China Physiological Signal Challenge (CPSC 2019):
    Challenging QRS Detection and Heart Rate Estimation from Single-Lead ECG Recordings
    """,
    about="""
    1. Training data consists of 2,000 single-lead ECG recordings collected from patients with cardiovascular disease (CVD)
    2. Each of the recording last for 10 s
    3. Sampling rate = 500 Hz
    4. Challenge official website [1]_.
    """,
    usage=[
        "ECG wave delineation",
    ],
    issues="""
    1. there're 13 records with unusual large values (> 20 mV):
        data_00098, data_00167, data_00173, data_00223, data_00224, data_00245, data_00813,
        data_00814, data_00815, data_00833, data_00841, data_00949, data_00950.

        .. code-block:: python

            >>> for rec in dr.all_records:
            >>>     data = dr.load_data(rec)
            >>>     if np.max(data) > 20:
            >>>         print(f"{rec} has max value ({np.max(data)} mV) > 20 mV")
            data_00173 has max value (32.72031811111111 mV) > 20 mV
            data_00223 has max value (32.75516713333333 mV) > 20 mV
            data_00224 has max value (32.7519272 mV) > 20 mV
            data_00245 has max value (32.75305293939394 mV) > 20 mV
            data_00813 has max value (32.75865595876289 mV) > 20 mV
            data_00814 has max value (32.75865595876289 mV) > 20 mV
            data_00815 has max value (32.75558282474227 mV) > 20 mV
            data_00833 has max value (32.76330123809524 mV) > 20 mV
            data_00841 has max value (32.727626558139534 mV) > 20 mV
            data_00949 has max value (32.75699667692308 mV) > 20 mV
            data_00950 has max value (32.769551661538465 mV) > 20 mV
    2. rpeak references (annotations) loaded from files has dtype = uint16, which would produce unexpected large positive values when subtracting values larger than it, rather than the correct negative value. This might cause confusion in computing metrics when using annotations subtracting (instead of being subtracted by) predictions.
    3. official scoring function has errors, which would falsely omit the interval between the 0-th and the 1-st ref rpeaks, thus potentially missing false positive
    """,
    references=[
        "http://2019.icbeb.org/Challenge.html",
    ],
    doi="10.1166/jmihi.2019.2800",
)


@add_docstring(_CPSC2019_INFO.format_database_docstring(), mode="prepend")
class CPSC2019(CPSCDataBase):
    """
    Parameters
    ----------
    db_dir : str or pathlib.Path, optional
        Storage path of the database.
        If not specified, data will be fetched from Physionet.
    working_dir : str, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Level of logging verbosity.
    kwargs : dict, optional
        Auxilliary key word arguments.

    """

    __name__ = "CPSC2019"

    def __init__(
        self,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[Union[str, Path]] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            db_name="cpsc2019",
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )

        self.fs = 500
        self.spacing = 1000 / self.fs

        self.rec_ext = "mat"
        self.ann_ext = "mat"

        # self.all_references = self.all_annotations
        self.rec_dir = self.db_dir / "data"
        self.ann_dir = self.db_dir / "ref"

        # aliases
        self.data_dir = self.rec_dir
        self.ref_dir = self.ann_dir

        self.n_records = 2000
        self._all_records = None
        self._all_annotations = None
        self._ls_rec()

    def _ls_rec(self) -> None:
        """Find all records in the database directory
        and store them (path, metadata, etc.) in some private attributes.
        """
        records_fn = self.db_dir / "records.json"
        self._all_records = [f"data_{i:05d}" for i in range(1, 1 + self.n_records)]
        self._all_annotations = [f"R_{i:05d}" for i in range(1, 1 + self.n_records)]
        self._df_records = pd.DataFrame()
        self.logger.info(
            "Please allow some time for the reader to confirm "
            "the existence of corresponding data files and annotation files..."
        )
        self._df_records["record"] = [
            f"data_{i:05d}" for i in range(1, 1 + self.n_records)
        ]
        self._df_records["path"] = self._df_records["record"].apply(
            lambda x: self.data_dir / x
        )
        self._df_records["annotation"] = self._df_records["record"].apply(
            lambda x: x.replace("data", "R")
        )
        self._df_records.index = self._df_records["record"]
        self._df_records = self._df_records.drop(columns="record")
        self._all_annotations = [f"R_{i:05d}" for i in range(1, 1 + self.n_records)]
        self._all_records = [
            rec
            for rec in self._all_records
            if self.get_absolute_path(rec, self.rec_ext).is_file()
        ]
        self._all_annotations = [
            ann
            for ann in self._all_annotations
            if self.get_absolute_path(ann, self.ann_ext, ann=True).is_file()
        ]
        common = set([rec.split("_")[1] for rec in self._all_records]) & set(
            [ann.split("_")[1] for ann in self._all_annotations]
        )
        if self._subsample is not None:
            # random subsample with ratio `self._subsample`
            size = min(len(common), max(1, int(round(self._subsample * len(common)))))
            common = DEFAULTS.RNG.choice(list(common), size, replace=False)
        common = sorted(common)
        self._all_records = [f"data_{item}" for item in common]
        self._all_annotations = [f"R_{item}" for item in common]
        self._df_records = self._df_records.loc[self._all_records]
        records_json = {"rec": self._all_records, "ann": self._all_annotations}
        records_fn.write_text(json.dumps(records_json, ensure_ascii=False))

    @property
    def all_annotations(self) -> List[str]:
        return self._all_annotations

    @property
    def all_references(self) -> List[str]:
        return self._all_annotations

    def get_subject_id(self, rec: Union[str, int]) -> int:
        """Attach a unique subject ID to the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        pid : int
            The ``subject_id`` corr. to `rec`.

        """
        if isinstance(rec, int):
            rec = self[rec]
        return int(f"19{int(rec.split('_')[1]):08d}")

    def get_absolute_path(
        self,
        rec: Union[str, int],
        extension: Optional[str] = None,
        ann: bool = False,
    ) -> Path:
        """Get the absolute path of the record `rec`.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        extension : str, optional
            Extension of the file.
            If not provided, no extension will be added.
        ann : bool, default False
            Whether to get the annotation file path or not.

        Returns
        -------
        Path
            Absolute path of the file.

        """
        if isinstance(rec, int):
            rec = self[rec]
        if extension is not None and not extension.startswith("."):
            extension = f".{extension}"
        if ann:
            rec = rec.replace("data", "R")
            return self.ann_dir / f"{rec}{extension or ''}"
        return self.data_dir / f"{rec}{extension or ''}"

    def load_data(
        self,
        rec: Union[int, str],
        data_format: str = "channel_first",
        units: str = "mV",
        fs: Optional[Real] = None,
    ) -> np.ndarray:
        """Load the ECG data of the record `rec`.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        data_format : str, default "channel_first"
            Format of the ECG data,
            "channel_last" (alias "lead_last"), or
            "channel_first" (alias "lead_first"), or
            "flat" (alias "plain").
        units : str or None, default "mV"
            Units of the output signal, can also be "μV" (with aliases "uV", "muV").
        fs : numbers.Real, optional
            If provided, the loaded data will be resampled to this frequency,
            otherwise the original sampling frequency will be used.

        Returns
        -------
        data : numpy.ndarray,
            The loaded ECG data.

        """
        fp = self.get_absolute_path(rec, self.rec_ext)
        data = loadmat(str(fp))["ecg"].astype(DEFAULTS.DTYPE.NP)
        if fs is not None and fs != self.fs:
            data = SS.resample_poly(data, fs, self.fs, axis=0).astype(data.dtype)
        if data_format.lower() in ["channel_first", "lead_first"]:
            data = data.T
        elif data_format.lower() in ["flat", "plain"]:
            data = data.flatten()
        elif data_format.lower() not in ["channel_last", "lead_last"]:
            raise ValueError(f"Invalid `data_format`: {data_format}")
        if units.lower() in ["uv", "μv", "muv"]:
            data = (1000 * data).astype(int)
        elif units.lower() != "mv":
            raise ValueError(f"Invalid `units`: {units}")
        return data

    def load_ann(self, rec: Union[int, str]) -> np.ndarray:
        """Load the annotations (indices of R peaks) of the record `rec`.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        ann : numpy.ndarray,
            Array of indices of R peaks.

        """
        fp = self.get_absolute_path(rec, self.ann_ext, ann=True)
        ann = loadmat(str(fp))["R_peak"].astype(int).flatten()
        return ann

    @add_docstring(load_ann.__doc__)
    def load_rpeaks(self, rec: Union[int, str]) -> np.ndarray:
        """
        alias of `self.load_ann`
        """
        return self.load_ann(rec=rec)

    @add_docstring(load_rpeaks.__doc__)
    def load_rpeak_indices(self, rec: Union[int, str]) -> np.ndarray:
        """
        alias of `self.load_rpeaks`
        """
        return self.load_rpeaks(rec=rec)

    def plot(
        self,
        rec: Union[int, str],
        data: Optional[np.ndarray] = None,
        ann: Optional[np.ndarray] = None,
        ticks_granularity: int = 0,
        **kwargs: Any,
    ) -> None:
        """Plot the ECG data of the record `rec`.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        data : numpy.ndarray, optional
            ECG signal to plot.
            If provided, data of `rec` will not be used,
            which is useful when plotting filtered data.
        ann : numpy.ndarray, optional
            Annotations (rpeak indices) for `data`.
            Ignored if `data` is None.
        ticks_granularity : int, default 0
            Granularity to plot axis ticks, the higher the more ticks.
            0 (no ticks) --> 1 (major ticks) --> 2 (major + minor ticks)

        """
        if isinstance(rec, int):
            rec = self[rec]
        if "plt" not in dir():
            import matplotlib.pyplot as plt

        if data is None:
            _data = self.load_data(rec, units="μV", data_format="flat")
        else:
            units = self._auto_infer_units(data)
            if units == "mV":
                _data = data * 1000
            elif units == "μV":
                _data = data.copy()

        duration = len(_data) / self.fs
        secs = np.linspace(0, duration, len(_data))
        if ann is None or data is None:
            rpeak_secs = self.load_rpeaks(rec) / self.fs
        else:
            rpeak_secs = np.array(ann) / self.fs

        fig_sz_w = int(DEFAULT_FIG_SIZE_PER_SEC * duration)
        y_range = np.max(np.abs(_data))
        fig_sz_h = 6 * y_range / 1500
        fig, ax = plt.subplots(figsize=(fig_sz_w, fig_sz_h))
        ax.plot(
            secs,
            _data,
            color="black",
            linewidth="2.0",
        )
        ax.axhline(y=0, linestyle="-", linewidth="1.0", color="red")
        if ticks_granularity >= 1:
            ax.xaxis.set_major_locator(plt.MultipleLocator(0.2))
            ax.yaxis.set_major_locator(plt.MultipleLocator(500))
            ax.grid(which="major", linestyle="-", linewidth="0.5", color="red")
        if ticks_granularity >= 2:
            ax.xaxis.set_minor_locator(plt.MultipleLocator(0.04))
            ax.yaxis.set_minor_locator(plt.MultipleLocator(100))
            ax.grid(which="minor", linestyle=":", linewidth="0.5", color="black")
        for r in rpeak_secs:
            ax.axvspan(r - 0.01, r + 0.01, color="green", alpha=0.9)
            ax.axvspan(r - 0.075, r + 0.075, color="green", alpha=0.3)
        ax.set_xlim(secs[0], secs[-1])
        ax.set_ylim(-y_range, y_range)
        ax.set_xlabel("Time [s]")
        ax.set_ylabel("Voltage [μV]")
        if kwargs.get("save_path", None):
            plt.savefig(kwargs["save_path"], dpi=200, bbox_inches="tight")
        else:
            plt.show()

    @property
    def url(self) -> str:
        return "https://www.dropbox.com/s/75nee0pqdy3f9r2/CPSC2019-train.zip?dl=1"

    @property
    def database_info(self) -> DataBaseInfo:
        return _CPSC2019_INFO

    @property
    def webpage(self) -> str:
        return "http://2019.icbeb.org/Challenge.html"


def compute_metrics(
    rpeaks_truths: Sequence[Union[np.ndarray, Sequence[int]]],
    rpeaks_preds: Sequence[Union[np.ndarray, Sequence[int]]],
    fs: Real,
    thr: float = 0.075,
    verbose: int = 0,
) -> float:
    """Metric (scoring) function modified from the official one,
    with errors fixed.

    Parameters
    ----------
    rpeaks_truths : sequence
        Sequence of ground truths of rpeaks locations from multiple records.
    rpeaks_preds : sequence
        Predictions of ground truths of rpeaks locations for multiple records.
    fs : numbers.Real
        Sampling frequency of ECG signal.
    thr : float, default 0.075
        Threshold for a prediction to be truth positive,
        with units in seconds.
    verbose : int, default 0
        Verbosity level for printing.

    Returns
    -------
    rec_acc : float
        Accuracy of predictions.

    """
    assert len(rpeaks_truths) == len(rpeaks_preds), (
        f"number of records does not match, truth indicates {len(rpeaks_truths)}, "
        f"while pred indicates {len(rpeaks_preds)}"
    )
    n_records = len(rpeaks_truths)
    record_flags = np.ones((len(rpeaks_truths),), dtype=float)
    thr_ = thr * fs
    if verbose >= 1:
        print(f"number of records = {n_records}")
        print(f"threshold in number of sample points = {thr_}")
    for idx, (truth_arr, pred_arr) in enumerate(zip(rpeaks_truths, rpeaks_preds)):
        false_negative = 0
        false_positive = 0
        true_positive = 0
        extended_truth_arr = np.concatenate((truth_arr.astype(int), [int(9.5 * fs)]))
        for j, t_ind in enumerate(extended_truth_arr[:-1]):
            next_t_ind = extended_truth_arr[j + 1]
            loc = np.where(np.abs(pred_arr - t_ind) <= thr_)[0]
            if j == 0:
                err = np.where(
                    (pred_arr >= 0.5 * fs + thr_) & (pred_arr <= t_ind - thr_)
                )[0]
            else:
                err = np.array([], dtype=int)
            err = np.append(
                err,
                np.where((pred_arr >= t_ind + thr_) & (pred_arr <= next_t_ind - thr_))[
                    0
                ],
            )

            false_positive += len(err)
            if len(loc) >= 1:
                true_positive += 1
                false_positive += len(loc) - 1
            elif len(loc) == 0:
                false_negative += 1

        if false_negative + false_positive > 1:
            record_flags[idx] = 0
        elif false_negative == 1 and false_positive == 0:
            record_flags[idx] = 0.3
        elif false_negative == 0 and false_positive == 1:
            record_flags[idx] = 0.7

        if verbose >= 2:
            print(
                f"for the {idx}-th record,\ntrue positive = {true_positive}\n"
                f"false positive = {false_positive}\nfalse negative = {false_negative}"
            )

    rec_acc = round(np.sum(record_flags) / n_records, 4)
    print(f"QRS_acc: {rec_acc}")
    print("Scoring complete.")

    return rec_acc



import json
import time
import warnings
from copy import deepcopy
from random import sample, shuffle
from typing import List, Optional, Sequence, Set, Tuple, Any

import numpy as np
import torch
from torch.utils.data.dataset import Dataset
from tqdm.auto import tqdm

from ...._preprocessors import PreprocManager
from ....cfg import CFG
from ....databases import CINC2020 as CR
from ....utils.misc import ReprMixin, list_sum
from ....utils.utils_data import ensure_siglen
from ....utils.utils_signal import remove_spikes_naive


__all__ = [
    "CINC2020Dataset",
]


class CINC2020Dataset(ReprMixin, Dataset):
    """Data generator for feeding data into pytorch models
    using the :class:`~torch_ecg.databases.CINC2020` database.

    Parameters
    ----------
    config : dict
        configurations for the :class:`Dataset`,
        ref. `CINC2020TrainCfg`.
        A simple example is as follows:

        .. code-block:: python

            >>> config = deepcopy(CINC2020TrainCfg)
            >>> config.db_dir = "some/path/to/db"
            >>> dataset = CINC2020Dataset(config, training=True, lazy=False)

    training : bool, default True
        If True, the training set will be loaded,
        otherwise the test (val) set will be loaded.
    lazy : bool, default True
        If True, the data will not be loaded immediately,
        instead, it will be loaded on demand.
    **reader_kwargs : dict, optional
        Keyword arguments for the database reader class.

    """

    __DEBUG__ = False
    __name__ = "CINC2020Dataset"

    def __init__(
        self,
        config: CFG,
        training: bool = True,
        lazy: bool = True,
        **reader_kwargs: Any,
    ) -> None:
        super().__init__()
        self.config = deepcopy(config)
        # assert self.config.db_dir is not None, "db_dir must be specified"
        if reader_kwargs.pop("db_dir", None) is not None:
            warnings.warn(
                "`db_dir` is specified in both config and reader_kwargs", RuntimeWarning
            )
        self.reader = CR(db_dir=self.config.db_dir, **reader_kwargs)
        self.config.db_dir = self.reader.db_dir
        self._TRANCHES = (
            self.config.tranche_classes.keys()
        )  # ["A", "B", "AB", "E", "F"]
        self.tranches = self.config.tranches_for_training
        self.training = training
        self.dtype = self.config.np_dtype
        assert not self.tranches or self.tranches in self._TRANCHES
        if self.tranches:
            self.all_classes = self.config.tranche_classes[self.tranches]
            self.class_weights = self.config.tranche_class_weights[self.tranches]
        else:
            self.all_classes = self.config.classes
            self.class_weights = self.config.class_weights
        self.config.all_classes = deepcopy(self.all_classes)
        self.n_classes = len(self.all_classes)
        # print(f"tranches = {self.tranches}, all_classes = {self.all_classes}")
        # print(f"class_weights = {dict_to_str(self.class_weights)}")
        cw = np.zeros((len(self.class_weights),), dtype=self.dtype)
        for idx, c in enumerate(self.all_classes):
            cw[idx] = self.class_weights[c]
        self.class_weights = torch.from_numpy(cw.astype(self.dtype)).view(
            1, self.n_classes
        )
        # validation also goes in batches, hence length has to be fixed
        self.siglen = self.config.input_len
        self.lazy = lazy

        self.records = self._train_test_split(
            self.config.train_ratio, force_recompute=False
        )
        # TODO: consider using `remove_spikes_naive` to treat these exceptional records
        self.records = [
            r for r in self.records if r not in self.reader.exceptional_records
        ]
        if self.__DEBUG__:
            self.records = sample(self.records, int(len(self.records) * 0.01))

        ppm_config = CFG(random=False)
        ppm_config.update(self.config)
        self.ppm = PreprocManager.from_config(ppm_config)
        # self.ppm.rearrange(["bandpass", "normalize"])

        self._signals = np.array([], dtype=self.dtype).reshape(
            0, len(self.config.leads), self.siglen
        )
        self._labels = np.array([], dtype=self.dtype).reshape(0, self.n_classes)
        if not self.lazy:
            self._load_all_data()

    def _load_all_data(self) -> None:
        """Load all data into memory."""
        fdr = _FastDataReader(self.reader, self.records, self.config, self.ppm)
        self._signals, self._labels = [], []
        with tqdm(
            range(len(fdr)),
            desc="Loading data",
            unit="record",
            dynamic_ncols=True,
            mininterval=1.0,
        ) as pbar:
            for idx in pbar:
                sig, lb = fdr[idx]
                self._signals.append(sig)
                self._labels.append(lb)
        self._signals = np.concatenate(self._signals, axis=0).astype(self.dtype)
        self._labels = np.concatenate(self._labels, axis=0)

    def _load_one_record(self, rec: str) -> Tuple[np.ndarray, np.ndarray]:
        """Load a record from the database using database reader.

        NOTE
        ----
        DO NOT USE THIS FUNCTION DIRECTLY for preloading data,
        use :class:`_FastDataReader` instead.

        Parameters
        ----------
        rec : str
            Name of the record to load.

        Returns
        -------
        values : numpy.ndarray
            The signal values of the record.
        labels : numpy.ndarray
            The labels of the record.

        """
        values = self.reader.load_resampled_data(
            rec, data_format=self.config.data_format, siglen=None
        )
        for idx in range(values.shape[0]):
            values[idx] = remove_spikes_naive(values[idx])
        values, _ = self.ppm(values, self.config.fs)
        values = ensure_siglen(
            values,
            siglen=self.siglen,
            fmt=self.config.data_format,
            tolerance=self.config.sig_slice_tol,
        ).astype(self.dtype)
        if values.ndim == 2:
            values = values[np.newaxis, ...]

        labels = self.reader.get_labels(rec, scored_only=True, fmt="a", normalize=True)
        labels = (
            np.isin(self.all_classes, labels)
            .astype(self.dtype)[np.newaxis, ...]
            .repeat(values.shape[0], axis=0)
        )

        return values, labels

    @property
    def signals(self) -> np.ndarray:
        """Cached signals, only available when `lazy=False`
        or preloading is performed manually.
        """
        return self._signals

    @property
    def labels(self) -> np.ndarray:
        """Cached labels, only available when `lazy=False`
        or preloading is performed manually.
        """
        return self._labels

    def __getitem__(self, index: int) -> Tuple[np.ndarray, np.ndarray]:
        return self.signals[index], self.labels[index]

    def __len__(self) -> int:
        return len(self._signals)

    def _train_test_split(
        self,
        train_ratio: float = 0.8,
        force_recompute: bool = False,
        force_valid: bool = False,
    ) -> List[str]:
        """Perform train-test split.

        It is ensured that both the train and the test set
        contain all classes if `force_valid` is True.

        Parameters
        ----------
        train_ratio : float, default 0.8
            Ratio of the train set in the whole dataset
            (or the whole tranche(s)).
        force_recompute : bool, default False
            If True, the train-test split will be recomputed,
            regardless of the existing ones stored in json files.
        force_valid : bool, default False
            If True, the train-test split will be recomputed
            if validity check fails.

        Returns
        -------
        records : List[str]
            List of the records split for training
            or for testing (validation).

        """
        time.sleep(1)
        start = time.time()
        print("\nstart performing train test split...\n")
        time.sleep(1)
        _TRANCHES = list("ABEF")
        _train_ratio = int(train_ratio * 100)
        _test_ratio = 100 - _train_ratio
        assert _train_ratio * _test_ratio > 0

        ns = "_ns" if len(self.config.special_classes) == 0 else ""
        file_suffix = f"_siglen_{self.siglen}{ns}.json"
        train_file = (
            self.reader.db_dir_base
            / f"{self.reader.db_name}_train_ratio_{_train_ratio}{file_suffix}"
        )
        test_file = (
            self.reader.db_dir_base
            / f"{self.reader.db_name}_test_ratio_{_test_ratio}{file_suffix}"
        )

        if force_recompute or not all([train_file.is_file(), test_file.is_file()]):
            tranche_records = {t: [] for t in _TRANCHES}
            train_set = {t: [] for t in _TRANCHES}
            test_set = {t: [] for t in _TRANCHES}
            for t in _TRANCHES:
                with tqdm(
                    self.reader.all_records[t],
                    total=len(self.reader.all_records[t]),
                    dynamic_ncols=True,
                    mininterval=1.0,
                ) as bar:
                    for rec in bar:
                        if rec in self.reader.exceptional_records:
                            # skip exceptional records
                            continue
                        rec_labels = self.reader.get_labels(
                            rec, scored_only=True, fmt="a", normalize=True
                        )
                        rec_labels = [
                            c for c in rec_labels if c in self.config.tranche_classes[t]
                        ]
                        if len(rec_labels) == 0:
                            # skip records with no scored class
                            continue
                        rec_samples = self.reader.load_resampled_data(rec).shape[1]
                        if rec_samples < self.siglen:
                            continue
                        tranche_records[t].append(rec)
                    print(
                        f"tranche {t} has {len(tranche_records[t])} valid records for training"
                    )
            for t in _TRANCHES:
                is_valid = False
                while not is_valid:
                    shuffle(tranche_records[t])
                    split_idx = int(len(tranche_records[t]) * train_ratio)
                    train_set[t] = tranche_records[t][:split_idx]
                    test_set[t] = tranche_records[t][split_idx:]
                    if force_valid:
                        is_valid = self._check_train_test_split_validity(
                            train_set[t],
                            test_set[t],
                            set(self.config.tranche_classes[t]),
                        )
                    else:
                        is_valid = True
            train_file.write_text(json.dumps(train_set, ensure_ascii=False))
            test_file.write_text(json.dumps(test_set, ensure_ascii=False))
        else:
            train_set = json.loads(train_file.read_text())
            test_set = json.loads(test_file.read_text())

        _tranches = list(self.tranches or "ABEF")
        if self.training:
            records = list_sum([train_set[k] for k in _tranches])
        else:
            records = list_sum([test_set[k] for k in _tranches])
        return records

    def _check_train_test_split_validity(
        self, train_set: List[str], test_set: List[str], all_classes: Set[str]
    ) -> bool:
        """Check if the train-test split is valid.

        The train-test split is valid iff
        records in both `train_set` and `test` contain all classes in `all_classes`

        Parameters
        ----------
        train_set : List[str]
            List of the records in the train set.
        test_set : List[str]
            List of the records in the test set.
        all_classes : Set[str]
            The set of all classes for training.

        Returns
        -------
        is_valid : bool
            Whether the train-test split is valid or not.

        """
        train_classes = set(
            list_sum([self.reader.get_labels(rec, fmt="a") for rec in train_set])
        )
        train_classes.intersection_update(all_classes)
        test_classes = set(
            list_sum([self.reader.get_labels(rec, fmt="a") for rec in test_set])
        )
        test_classes.intersection_update(all_classes)
        is_valid = len(all_classes) == len(train_classes) == len(test_classes)
        print(
            f"all_classes = {all_classes}\n"
            f"train_classes = {train_classes}\ntest_classes = {test_classes}\n"
            f"is_valid = {is_valid}"
        )
        return is_valid

    def persistence(self) -> None:
        """Save the processed dataset to disk."""
        _TRANCHES = "ABEF"
        if self.training:
            ratio = int(self.config.train_ratio * 100)
        else:
            ratio = 100 - int(self.config.train_ratio * 100)
        fn_suffix = f"tranches_{self.tranches or _TRANCHES}_ratio_{ratio}"
        if self.config.bandpass is not None:
            bp_low = max(0, self.config.bandpass[0])
            bp_high = min(self.config.bandpass[1], self.config.fs // 2)
            fn_suffix = fn_suffix + f"_bp_{bp_low:.1f}_{bp_high:.1f}"
        fn_suffix = fn_suffix + f"_siglen_{self.siglen}"

        X, y = [], []
        with tqdm(
            range(self.__len__()),
            total=self.__len__(),
            dynamic_ncols=True,
            mininterval=1.0,
        ) as bar:
            for idx in bar:
                values, labels = self.__getitem__(idx)
                X.append(values)
                y.append(labels)
        X, y = np.array(X), np.array(y)
        print(f"X.shape = {X.shape}, y.shape = {y.shape}")
        filename = f"{'train' if self.training else 'test'}_X_{fn_suffix}.npy"
        np.save(self.reader.db_dir_base / filename, X)
        print(f"X saved to {filename}")
        filename = f"{'train' if self.training else 'test'}_y_{fn_suffix}.npy"
        np.save(self.reader.db_dir_base / filename, y)
        print(f"y saved to {filename}")

    def _check_nan(self) -> None:
        """Check if there are nan values in the dataset."""
        for idx, (values, labels) in enumerate(self):
            if np.isnan(values).any():
                print(f"values of {self.records[idx]} have nan values")
            if np.isnan(labels).any():
                print(f"labels of {self.records[idx]} have nan values")

    def extra_repr_keys(self) -> List[str]:
        return [
            "training",
            "tranches",
            "reader",
        ]


class _FastDataReader(ReprMixin, Dataset):
    """Fast data reader.

    Parameters
    ----------
    reader : CR
        The reader to read the data.
    records : Sequence[str]
        The list of records to read.
    config : CFG
        The configuration.
    ppm : PreprocManager, optional
        The preprocessor manager.

    """

    def __init__(
        self,
        reader: CR,
        records: Sequence[str],
        config: CFG,
        ppm: Optional[PreprocManager] = None,
    ) -> None:
        self.reader = reader
        self.records = records
        self.config = config
        self.ppm = ppm
        self.dtype = self.config.np_dtype

    def __len__(self) -> int:
        return len(self.records)

    def __getitem__(self, index: int) -> Tuple[np.ndarray, np.ndarray]:
        rec = self.records[index]
        values = self.reader.load_resampled_data(
            rec, data_format=self.config.data_format, siglen=None
        )
        for idx in range(values.shape[0]):
            values[idx] = remove_spikes_naive(values[idx])
        if self.ppm:
            values, _ = self.ppm(values, self.config.fs)
        values = ensure_siglen(
            values,
            siglen=self.config.input_len,
            fmt=self.config.data_format,
            tolerance=self.config.sig_slice_tol,
        ).astype(self.dtype)
        if values.ndim == 2:
            values = values[np.newaxis, ...]

        labels = self.reader.get_labels(rec, scored_only=True, fmt="a", normalize=True)
        labels = (
            np.isin(self.config.all_classes, labels)
            .astype(self.dtype)[np.newaxis, ...]
            .repeat(values.shape[0], axis=0)
        )

        return values, labels

    def extra_repr_keys(self) -> List[str]:
        return [
            "reader",
            "ppm",
        ]



"""
configurations for signal preprocess, feature extraction, training, etc.
along with some constants

"Brady", "LAD", "RAD", "PR", "LQRSV" are treated exceptionally, as special classes

"""

from copy import deepcopy
from typing import List

from ....cfg import CFG, DEFAULTS
from ....utils import ecg_arrhythmia_knowledge as EAK
from ...aux_data.cinc2020_aux_data import get_class_weight

__all__ = [
    "CINC2020TrainCfg",
]


# special classes using special detectors
# _SPECIAL_CLASSES = ["Brady", "LAD", "RAD", "PR", "LQRSV"]
_SPECIAL_CLASSES = []
_NAME = "cinc2020"


def _assign_classes(cfg: CFG, special_classes: List[str]) -> None:
    """ """
    cfg.special_classes = deepcopy(special_classes)
    cfg.tranche_class_weights = CFG(
        {
            t: get_class_weight(
                t,
                exclude_classes=cfg.special_classes,
                scored_only=True,
                threshold=20,
                min_weight=cfg.min_class_weight,
            )
            for t in [
                "A",
                "B",
                "AB",
                "E",
                "F",
            ]
        }
    )
    cfg.tranche_classes = CFG(
        {t: sorted(list(t_cw.keys())) for t, t_cw in cfg.tranche_class_weights.items()}
    )

    cfg.class_weights = get_class_weight(
        tranches="ABEF",
        exclude_classes=cfg.special_classes,
        scored_only=True,
        threshold=20,
        min_weight=cfg.min_class_weight,
    )
    cfg.classes = sorted(list(cfg.class_weights.keys()))


# training configurations for machine learning and deep learning
CINC2020TrainCfg = CFG()
CINC2020TrainCfg.torch_dtype = DEFAULTS.DTYPE.TORCH
CINC2020TrainCfg.np_dtype = DEFAULTS.DTYPE.NP

# configs of files
CINC2020TrainCfg.db_dir = None
CINC2020TrainCfg.log_dir = DEFAULTS.log_dir / _NAME
CINC2020TrainCfg.log_dir.mkdir(parents=True, exist_ok=True)
CINC2020TrainCfg.checkpoints = DEFAULTS.checkpoints / _NAME
CINC2020TrainCfg.checkpoints.mkdir(parents=True, exist_ok=True)
CINC2020TrainCfg.model_dir = DEFAULTS.model_dir / _NAME
CINC2020TrainCfg.model_dir.mkdir(parents=True, exist_ok=True)

CINC2020TrainCfg.final_model_name = None
CINC2020TrainCfg.keep_checkpoint_max = 20

CINC2020TrainCfg.leads = deepcopy(EAK.Standard12Leads)

# configs of training data
CINC2020TrainCfg.fs = 500
CINC2020TrainCfg.data_format = "channel_first"

CINC2020TrainCfg.train_ratio = 0.8
CINC2020TrainCfg.min_class_weight = 0.5
CINC2020TrainCfg.tranches_for_training = ""  # one of "", "AB", "E", "F", "G"

# assign classes, class weights, tranche classes, etc.
_assign_classes(CINC2020TrainCfg, _SPECIAL_CLASSES)

# configs of signal preprocessing
CINC2020TrainCfg.normalize = CFG(
    method="z-score",
    mean=0.0,
    std=1.0,
)
# frequency band of the filter to apply, should be chosen very carefully
CINC2020TrainCfg.bandpass = None
# CINC2020TrainCfg.bandpass = CFG(
#     lowcut=0.5,
#     highcut=60,
# )

# configs of data aumentation
# CINC2020TrainCfg.label_smooth = CFG(
#     prob=0.8,
#     smoothing=0.1,
# )
CINC2020TrainCfg.label_smooth = False
CINC2020TrainCfg.random_masking = False
CINC2020TrainCfg.stretch_compress = False  # stretch or compress in time axis
CINC2020TrainCfg.mixup = CFG(
    prob=0.6,
    alpha=0.3,
)

# configs of training epochs, batch, etc.
CINC2020TrainCfg.n_epochs = 50
# TODO: automatic adjust batch size according to GPU capacity
# https://stackoverflow.com/questions/45132809/how-to-select-batch-size-automatically-to-fit-gpu
CINC2020TrainCfg.batch_size = 64
# CINC2020TrainCfg.max_batches = 500500

# configs of optimizers and lr_schedulers
CINC2020TrainCfg.optimizer = "adamw_amsgrad"  # "sgd", "adam", "adamw"
CINC2020TrainCfg.momentum = 0.949  # default values for corresponding PyTorch optimizers
CINC2020TrainCfg.betas = (
    0.9,
    0.999,
)  # default values for corresponding PyTorch optimizers
CINC2020TrainCfg.decay = 1e-2  # default values for corresponding PyTorch optimizers

CINC2020TrainCfg.learning_rate = 1e-4  # 1e-3
CINC2020TrainCfg.lr = CINC2020TrainCfg.learning_rate

CINC2020TrainCfg.lr_scheduler = (
    "one_cycle"  # "one_cycle", "plateau", "burn_in", "step", None
)
CINC2020TrainCfg.lr_step_size = 50
CINC2020TrainCfg.lr_gamma = 0.1
CINC2020TrainCfg.max_lr = 2e-3  # for "one_cycle" scheduler, to adjust via expriments

CINC2020TrainCfg.burn_in = 400
CINC2020TrainCfg.steps = [5000, 10000]

CINC2020TrainCfg.early_stopping = CFG()  # early stopping according to challenge metric
CINC2020TrainCfg.early_stopping.min_delta = 0.001  # should be non-negative
CINC2020TrainCfg.early_stopping.patience = 10

# configs of loss function
# CINC2020TrainCfg.loss = "BCEWithLogitsLoss"
# CINC2020TrainCfg.loss = "BCEWithLogitsWithClassWeightLoss"
CINC2020TrainCfg.loss = "AsymmetricLoss"  # "FocalLoss"
CINC2020TrainCfg.loss_kw = CFG(gamma_pos=0, gamma_neg=0.2, implementation="deep-psp")
CINC2020TrainCfg.flooding_level = (
    0.0  # flooding performed if positive, typically 0.45-0.55 for cinc2020?
)

CINC2020TrainCfg.monitor = "challenge_metric"

CINC2020TrainCfg.log_step = 20
CINC2020TrainCfg.eval_every = 20

# configs of model selection
# "resnet_nature_comm_se", "multi_scopic_leadwise", "vgg16", "vgg16_leadwise",
CINC2020TrainCfg.cnn_name = "resnet_nature_comm_bottle_neck_se"
CINC2020TrainCfg.rnn_name = "none"  # "none", "lstm"
CINC2020TrainCfg.attn_name = "none"  # "none", "se", "gc", "nl"

# configs of inputs and outputs
# almost all records have duration >= 8s, most have duration >= 10s
# use `utils.utils_signal.ensure_siglen` to ensure signal length
CINC2020TrainCfg.input_len = int(500 * 10.0)
# tolerance for records with length shorter than `CINC2020TrainCfg.input_len`
CINC2020TrainCfg.input_len_tol = int(0.2 * CINC2020TrainCfg.input_len)
CINC2020TrainCfg.sig_slice_tol = 0.4  # None, do no slicing
CINC2020TrainCfg.siglen = CINC2020TrainCfg.input_len


# constants for model inference
CINC2020TrainCfg.bin_pred_thr = 0.5
# `bin_pred_look_again_tol` is used when no prob is greater than `bin_pred_thr`,
# then the prediction would be the one with the highest prob.,
# along with those with prob. no less than the highest prob. minus `bin_pred_look_again_tol`
CINC2020TrainCfg.bin_pred_look_again_tol = 0.03
CINC2020TrainCfg.bin_pred_nsr_thr = 0.1



# -*- coding: utf-8 -*-

import math
import time
import warnings
from pathlib import Path
from typing import Any, Optional, Sequence, Union, List

import numpy as np
import pandas as pd

from ...cfg import DEFAULTS
from ...utils.misc import get_record_list_recursive3, add_docstring
from ..base import DEFAULT_FIG_SIZE_PER_SEC, PhysioNetDataBase, DataBaseInfo


__all__ = [
    "CINC2017",
]


_CINC2017_INFO = DataBaseInfo(
    title="""
    AF Classification from a Short Single Lead ECG Recording
    -- The PhysioNet Computing in Cardiology Challenge 2017
    """,
    about="""
    1. training set contains 8,528 single lead ECG recordings lasting from 9 s to just over 60 s, and the test set contains 3,658 ECG recordings of similar lengths
    2. records are of frequency 300 Hz and have been band pass filtered
    3. data distribution:

        +------------------+--------------+-------------------------------------+
        |                  |              |           Time length (s)           |
        |  Type            | # recording  +------+------+------+--------+-------+
        |                  |              | Mean | SD   | Max  | Median | Min   |
        +==================+==============+======+======+======+========+=======+
        | Normal           | 5154         | 31.9 | 10.0 | 61.0 | 30     | 9.0   |
        +------------------+--------------+------+------+------+--------+-------+
        | AF               | 771          | 31.6 | 12.5 | 60   | 30     | 10.0  |
        +------------------+--------------+------+------+------+--------+-------+
        | Other rhythm     | 2557         | 34.1 | 11.8 | 60.9 | 30     | 9.1   |
        +------------------+--------------+------+------+------+--------+-------+
        | Noisy            | 46           | 27.1 | 9.0  | 60   | 30     | 10.2  |
        +------------------+--------------+------+------+------+--------+-------+
        | Total            | 8528         | 32.5 | 10.9 | 61.0 | 30     | 9.0   |
        +------------------+--------------+------+------+------+--------+-------+

    4. Webpage of the database on PhysioNet [1]_.
    """,
    usage=[
        "Atrial fibrillation (AF) detection",
    ],
    references=[
        "https://physionet.org/content/challenge-2017/",
    ],
    doi=[
        "10.22489/CinC.2017.065-469",
    ],
)


@add_docstring(_CINC2017_INFO.format_database_docstring(), mode="prepend")
class CINC2017(PhysioNetDataBase):
    """
    Parameters
    ----------
    db_dir : str or pathlib.Path, optional
        Storage path of the database.
        If not specified, data will be fetched from Physionet.
    working_dir : str, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Level of logging verbosity.
    kwargs : dict, optional
        Auxilliary key word arguments.

    """

    __name__ = "CINC2017"

    def __init__(
        self,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[Union[str, Path]] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            db_name="challenge-2017",
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )

        self.fs = 300

        self.rec_ext = "mat"
        self.ann_ext = "hea"

        self._all_records = []
        self._df_ann = pd.DataFrame()
        self._df_ann_ori = pd.DataFrame()
        self._all_ann = []
        self._ls_rec()

        self.d_ann_names = {
            "N": "Normal rhythm",
            "A": "AF rhythm",
            "O": "Other rhythm",
            "~": "Noisy",
        }
        self.palette = {
            "N": "green",
            "A": "red",
            "O": "yellow",
            "~": "blue",
        }

        # self._url_compressed = (
        #     "https://physionet.org/static/published-projects/challenge-2017/"
        #     "af-classification-from-a-short-single-lead-ecg-recording-"
        #     "the-physionetcomputing-in-cardiology-challenge-2017-1.0.0.zip"
        # )
        self._url_compressed = self.get_file_download_url("training2017.zip")

    def _ls_rec(self) -> None:
        """Find all records in the database directory
        and store them (path, metadata, etc.) in some private attributes.
        """
        record_list_fp = self.db_dir / "RECORDS"
        self._df_records = pd.DataFrame()
        if record_list_fp.is_file():
            self._df_records["record"] = [
                item
                for item in record_list_fp.read_text().splitlines()
                if len(item) > 0
            ]
            if len(self._df_records) > 0:
                if self._subsample is not None:
                    size = min(
                        len(self._df_records),
                        max(1, int(round(self._subsample * len(self._df_records)))),
                    )
                    self._df_records = self._df_records.sample(
                        n=size, random_state=DEFAULTS.SEED, replace=False
                    )
                self._df_records["path"] = self._df_records["record"].apply(
                    lambda x: (self.db_dir / x).resolve()
                )
                self._df_records = self._df_records[
                    self._df_records["path"].apply(lambda x: x.is_file())
                ]
                self._df_records["record"] = self._df_records["path"].apply(
                    lambda x: x.name
                )

        if len(self._df_records) == 0:
            self.logger.info(
                "Please wait patiently to let the reader find "
                "all records of the database from local storage..."
            )
            start = time.time()
            self._df_records["path"] = get_record_list_recursive3(
                db_dir=str(self.db_dir),
                rec_patterns=f"A[\\d]{{5}}\\.{self.rec_ext}",
                relative=False,
            )
            if self._subsample is not None:
                size = min(
                    len(self._df_records),
                    max(1, int(round(self._subsample * len(self._df_records)))),
                )
                self._df_records = self._df_records.sample(
                    n=size, random_state=DEFAULTS.SEED, replace=False
                )
            self._df_records["path"] = self._df_records["path"].apply(lambda x: Path(x))
            self.logger.info(f"Done in {time.time() - start:.3f} seconds!")
            self._df_records["record"] = self._df_records["path"].apply(
                lambda x: x.name
            )
        self._df_records.set_index("record", inplace=True)
        self._all_records = self._df_records.index.values.tolist()

        ann_file = list(self.db_dir.rglob("REFERENCE.csv"))
        if len(ann_file) > 0:
            self._df_ann = pd.read_csv(ann_file[0], header=None)
            self._df_ann.columns = ["rec", "ann"]
        else:
            self._df_ann = pd.DataFrame(columns=["rec", "ann"])
            warnings.warn(
                "Cannot find the annotation file `REFERENCE.csv`!",
                RuntimeWarning,
            )
        ann_file = list(self.db_dir.rglob("REFERENCE-original.csv"))
        if len(ann_file) > 0:
            self._df_ann_ori = pd.read_csv(ann_file[0], header=None)
            self._df_ann_ori.columns = ["rec", "ann"]
        else:
            self._df_ann_ori = pd.DataFrame(columns=["rec", "ann"])
            warnings.warn(
                "Cannot find the annotation file `REFERENCE-original.csv`!",
                RuntimeWarning,
            )
        # ["N", "A", "O", "~"]
        self._all_ann = list(
            set(
                self._df_ann.ann.unique().tolist()
                + self._df_ann_ori.ann.unique().tolist()
            )
        )

    def load_ann(
        self, rec: Union[str, int], original: bool = False, ann_format: str = "a"
    ) -> str:
        """Load the annotation of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        original : bool, default False
            If True, load annotations from
            the annotation file ``REFERENCE-original.csv``,
            otherwise from ``REFERENCE.csv``.
        ann_format : {"a", "f"}, optional
            Format of returned annotation, by default "a".

                - "a" - abbreviation
                - "f" - full name

        Returns
        -------
        ann : str
            Annotation (label) of the record.

        """
        if isinstance(rec, int):
            rec = self[rec]
        assert rec in self.all_records and ann_format.lower() in ["a", "f"]
        if original:
            df = self._df_ann_ori
        else:
            df = self._df_ann
        row = df[df.rec == rec].iloc[0]
        ann = row.ann
        if ann_format.lower() == "f":
            ann = self.d_ann_names[ann]
        return ann

    def plot(
        self,
        rec: Union[str, int],
        data: Optional[np.ndarray] = None,
        ann: Optional[str] = None,
        ticks_granularity: int = 0,
        rpeak_inds: Optional[Union[Sequence[int], np.ndarray]] = None,
    ) -> None:
        """Plot the ECG signal of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        data : numpy.ndarray, optional
            The ECG signal to plot.
            If not None, data of `rec` will not be used.
            This is useful when plotting filtered data.
        ann : dict, optional,
            Annotations for `data`, which is a dict with keys
            "SPB_indices", "PVC_indices",
            and with :class:`~numpy.ndarray` values.
            Ignored if `data` is None.
        ticks_granularity : int, default 0
            Granularity to plot axis ticks, the higher the more ticks.
            0 (no ticks) --> 1 (major ticks) --> 2 (major + minor ticks)
        rpeak_inds : array_like, optional
            Array of indices of R peaks.

        Returns
        -------
        None

        """
        if isinstance(rec, int):
            rec = self[rec]
        if "plt" not in dir():
            import matplotlib.pyplot as plt
        import matplotlib.patches as mpatches

        if data is None:
            _data = self.load_data(
                rec,
                units="μV",
                data_format="flat",
            )
        else:
            units = self._auto_infer_units(data)
            if units == "mV":
                _data = data * 1000
            elif units == "μV":
                _data = data.copy()

        if ann is None or data is None:
            ann = self.load_ann(rec, ann_format="a")
            ann_fullname = self.load_ann(rec, ann_format="f")
        else:
            ann_fullname = self.d_ann_names.get(ann, ann)
        patch = mpatches.Patch(color=self.palette.get(ann, "blue"), label=ann_fullname)

        if rpeak_inds is not None:
            rpeak_secs = np.array(rpeak_inds) / self.fs

        line_len = self.fs * 25  # 25 seconds
        nb_lines = math.ceil(len(_data) / line_len)

        for idx in range(nb_lines):
            seg = _data[idx * line_len : (idx + 1) * line_len]
            secs = (np.arange(len(seg)) + idx * line_len) / self.fs
            fig_sz_w = int(round(DEFAULT_FIG_SIZE_PER_SEC * len(seg) / self.fs))
            y_range = np.max(np.abs(seg)) + 100
            fig_sz_h = 6 * y_range / 1500
            fig, ax = plt.subplots(figsize=(fig_sz_w, fig_sz_h))
            ax.plot(secs, seg, color="black")
            ax.axhline(y=0, linestyle="-", linewidth="1.0", color="red")
            if ticks_granularity >= 1:
                ax.xaxis.set_major_locator(plt.MultipleLocator(0.2))
                ax.yaxis.set_major_locator(plt.MultipleLocator(500))
                ax.grid(which="major", linestyle="-", linewidth="0.5", color="red")
            if ticks_granularity >= 2:
                ax.xaxis.set_minor_locator(plt.MultipleLocator(0.04))
                ax.yaxis.set_minor_locator(plt.MultipleLocator(100))
                ax.grid(which="minor", linestyle=":", linewidth="0.5", color="black")
            ax.legend(handles=[patch], loc="lower left", prop={"size": 16})
            if rpeak_inds is not None:
                for r in rpeak_secs:
                    ax.axvspan(r - 0.01, r + 0.01, color="green", alpha=0.7)
            ax.set_xlim(secs[0], secs[-1])
            ax.set_ylim(-y_range, y_range)
            ax.set_xlabel("Time [s]")
            ax.set_ylabel("Voltage [μV]")
            plt.show()

    @property
    def _validation_set(self) -> List[str]:
        """The validation set specified at
        https://physionet.org/content/challenge-2017/1.0.0/
        """
        return (
            "A00001,A00002,A00003,A00004,A00005,A00006,A00007,A00008,A00009,A00010,"
            "A00011,A00012,A00013,A00014,A00015,A00016,A00017,A00018,A00019,A00020,"
            "A00021,A00022,A00023,A00024,A00025,A00026,A00027,A00028,A00029,A00030,"
            "A00031,A00032,A00033,A00034,A00035,A00036,A00037,A00038,A00039,A00040,"
            "A00041,A00042,A00043,A00044,A00045,A00046,A00047,A00048,A00049,A00050,"
            "A00051,A00052,A00053,A00054,A00055,A00056,A00057,A00058,A00059,A00060,"
            "A00061,A00062,A00063,A00064,A00065,A00066,A00067,A00068,A00069,A00070,"
            "A00071,A00072,A00073,A00074,A00075,A00076,A00077,A00078,A00079,A00080,"
            "A00081,A00082,A00083,A00084,A00085,A00086,A00087,A00088,A00089,A00090,"
            "A00091,A00092,A00093,A00094,A00095,A00096,A00097,A00098,A00099,A00100,"
            "A00101,A00102,A00103,A00104,A00105,A00106,A00107,A00108,A00109,A00110,"
            "A00111,A00112,A00113,A00114,A00115,A00116,A00117,A00118,A00119,A00120,"
            "A00121,A00122,A00123,A00124,A00125,A00126,A00127,A00128,A00129,A00130,"
            "A00131,A00132,A00133,A00134,A00135,A00136,A00137,A00138,A00139,A00140,"
            "A00141,A00142,A00143,A00144,A00145,A00146,A00147,A00148,A00149,A00150,"
            "A00151,A00152,A00153,A00154,A00155,A00156,A00157,A00158,A00159,A00160,"
            "A00161,A00162,A00163,A00164,A00165,A00166,A00167,A00168,A00169,A00170,"
            "A00171,A00172,A00173,A00174,A00175,A00176,A00177,A00178,A00179,A00180,"
            "A00181,A00182,A00183,A00184,A00185,A00186,A00187,A00188,A00189,A00190,"
            "A00191,A00192,A00193,A00194,A00195,A00196,A00197,A00198,A00199,A00200,"
            "A00201,A00202,A00203,A00204,A00205,A00206,A00207,A00208,A00209,A00210,"
            "A00211,A00212,A00213,A00214,A00215,A00216,A00217,A00218,A00219,A00220,"
            "A00221,A00222,A00223,A00224,A00225,A00226,A00227,A00228,A00229,A00230,"
            "A00231,A00232,A00233,A00234,A00235,A00236,A00237,A00238,A00239,A00240,"
            "A00241,A00242,A00244,A00245,A00247,A00248,A00249,A00253,A00267,A00271,"
            "A00301,A00321,A00375,A00395,A00397,A00405,A00422,A00432,A00438,A00439,"
            "A00441,A00456,A00465,A00473,A00486,A00509,A00519,A00520,A00524,A00542,"
            "A00551,A00585,A01006,A01070,A01246,A01299,A01521,A01567,A01707,A01727,"
            "A01772,A01833,A02168,A02372,A02772,A02785,A02833,A03549,A03738,A04086,"
            "A04137,A04170,A04186,A04216,A04282,A04452,A04522,A04701,A04735,A04805"
        ).split(",")

    @property
    def database_info(self) -> DataBaseInfo:
        return _CINC2017_INFO



"""
"""

import numpy as np
import torch
import pytest

from torch_ecg.cfg import CFG, DEFAULTS, DTYPE


def test_cfg():
    c = CFG(hehe={"a": 1, "b": 2})
    c.update(hehe={"a": [-1]})
    assert str(c) == repr(c) == r"{'hehe': {'a': [-1], 'b': 2}}"
    c.update(hehe={"c": -10})
    assert str(c) == repr(c) == r"{'hehe': {'a': [-1], 'b': 2, 'c': -10}}"
    assert c.hehe.pop("a") == [-1]
    assert str(c) == repr(c) == r"{'hehe': {'b': 2, 'c': -10}}"

    c = CFG({1: 2, 3: 4})
    assert str(c) == repr(c) == r"{1: 2, 3: 4}"

    with pytest.raises(TypeError, match="expected at most 1 arguments, got 2"):
        CFG(1, 2)


def test_dtype():
    dtp = DTYPE("float32")
    assert (
        str(dtp)
        == repr(dtp)
        == "DTYPE(STR='float32', NP=dtype('float32'), TORCH=torch.float32, INT=32)"
    )

    with pytest.raises(TypeError, match="data type 'hehe' not understood"):
        DTYPE("hehe")

    with pytest.raises(AssertionError, match="inconsistent dtype"):
        DTYPE("float32", INT=64)


def test_defaults():
    assert DEFAULTS.DTYPE == DTYPE("float32")
    assert DEFAULTS.dtype == torch.float32
    DEFAULTS.change_dtype(torch.float16)
    assert DEFAULTS.dtype == torch.float16
    DEFAULTS.change_dtype(np.float32)
    assert DEFAULTS.dtype == torch.float32

    with pytest.raises(
        TypeError, match="`dtype` must be a str or np.dtype or torch.dtype"
    ):
        DEFAULTS.change_dtype(32)
    with pytest.raises(AssertionError, match="`dtype` must be one of "):
        DEFAULTS.change_dtype("float128")

    assert DEFAULTS.SEED == 42
    DEFAULTS.set_seed(100)
    assert DEFAULTS.SEED == 100



"""
"""

import inspect
import math
from copy import deepcopy
from abc import ABC, abstractmethod
from typing import Union, List, Sequence, Tuple

import numpy as np
import torch
from torch.nn import functional as F
from einops.layers.torch import Rearrange

from ..cfg import CFG, DEFAULTS
from ..utils.misc import ReprMixin, add_docstring
from ..utils.utils_nn import compute_conv_output_shape
from ..utils.utils_signal_t import Spectrogram


__all__ = [
    "InputConfig",
    "WaveformInput",
    "FFTInput",
    "SpectrogramInput",
]


class InputConfig(CFG):
    """A Class to store the configuration of the input.

    Parameters
    ----------
    input_type : {"waveform", "fft", "spectrogram"}, optional
        Type of the input.
    n_channels : int
        Number of channels of the input.
    n_samples : int
        Number of samples of the input.
    ensure_batch_dim : bool
        Whether to ensure the transformed input has a batch dimension.

    Examples
    --------
    .. code-block:: python

        input_config = InputConfig(
            input_type="waveform",
            n_channels=12,
            n_samples=5000,
        )

    """

    __name__ = "InputConfig"

    def __init__(
        self,
        *args: Union[CFG, dict],
        input_type: str,
        n_channels: int,
        n_samples: int = -1,
        ensure_batch_dim: bool = True,
        **kwargs: dict,
    ) -> None:
        super().__init__(
            *args,
            input_type=input_type,
            n_channels=n_channels,
            n_samples=n_samples,
            ensure_batch_dim=ensure_batch_dim,
            **kwargs,
        )
        assert (
            "n_channels" in self and self.n_channels > 0
        ), f"`n_channels` must be positive, got {self.n_channels}"
        assert "n_samples" in self and (
            self.n_samples > 0 or self.n_samples == -1
        ), f"`n_samples` must be positive or -1, got {self.n_samples}"
        assert "input_type" in self and self.input_type.lower() in [
            "waveform",
            "fft",
            "spectrogram",
        ], f"`input_type` must be one of ['waveform', 'fft', 'spectrogram'], got {self.input_type}"
        self.input_type = self.input_type.lower()
        if self.input_type in [
            "spectrogram",
        ]:
            assert (
                "n_bins" in self
            ), f"`n_bins` must be specified for {self.input_type} input"
            assert (
                "fs" in self or "sample_rate" in self
            ), f"`fs` or `sample_rate` must be specified for {self.input_type} input"


class BaseInput(ReprMixin, ABC):
    """Base class for all input classes.

    Parameters
    ----------
    config : InputConfig
        The configuration of the input.

    """

    __name__ = "BaseInput"

    def __init__(self, config: InputConfig) -> None:
        """ """
        assert isinstance(
            config, InputConfig
        ), "`config` must be an instance of `InputConfig`"
        self._config = deepcopy(config)
        self._values = None
        self._dtype = self._config.get("dtype", DEFAULTS.DTYPE.TORCH)
        self._device = self._config.get("device", DEFAULTS.device)
        self._post_init()

    def __call__(self, waveform: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        """Method to transform the waveform to the input tensor.

        Parameters
        ----------
        waveform : numpy.ndarray or torch.Tensor
            The waveform to be transformed.

        Returns
        -------
        torch.Tensor
            The transformed waveform.

        """
        return self.from_waveform(waveform)

    @abstractmethod
    def _from_waveform(self, waveform: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        """Internal method to convert the waveform to the input tensor."""
        raise NotImplementedError

    def from_waveform(self, waveform: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        """Transform the waveform to the input tensor.

        Parameters
        ----------
        waveform : numpy.ndarray or torch.Tensor
            The waveform to be transformed.

        Returns
        -------
        torch.Tensor
            The transformed waveform.

        """
        assert waveform.shape[-2:] == (self.n_channels, self.n_samples,), (
            f"`waveform` shape must be `(batch_size, {self.n_channels}, {self.n_samples})` "
            f"or `({self.n_channels}, {self.n_samples})`, got `{waveform.shape}`"
        )
        input_tensors = self._from_waveform(waveform)
        if waveform.ndim == 2 and self._config.ensure_batch_dim:
            input_tensors = input_tensors.unsqueeze(0)
        return input_tensors

    @abstractmethod
    def _post_init(self) -> None:
        """Method to be called after initialization"""
        raise NotImplementedError

    @property
    def values(self) -> torch.Tensor:
        return self._values

    @property
    def n_channels(self) -> int:
        return self._config.n_channels

    @property
    def n_samples(self) -> int:
        return self._config.n_samples

    @property
    def input_channels(self) -> int:
        channel_dim = {
            "waveform": -2,
            "fft": -2,
            # "spectrogram": -3,  # implemented in `SpectrogramInput`
        }
        if self.values is not None:
            return self.values.shape[channel_dim[self.input_type]]
        return self.compute_input_shape((self.n_channels, self.n_samples))[
            channel_dim[self.input_type]
        ]

    @property
    def input_samples(self) -> int:
        if self.values is not None:
            return self.values.shape[-1]
        return self.compute_input_shape((self.n_channels, self.n_samples))[-1]

    @property
    def input_type(self) -> str:
        return self._config.input_type

    @property
    def dtype(self) -> torch.dtype:
        return self._dtype

    @property
    def device(self) -> torch.device:
        return self._device

    def compute_input_shape(
        self, waveform_shape: Union[Sequence[int], torch.Size]
    ) -> Tuple[Union[type(None), int], ...]:
        """Computes the input shape of the model based on
        the input type and the waveform shape.

        Parameters
        ----------
        waveform_shape : Sequence[int] or torch.Size
            The shape of the waveform.

        Returns
        -------
        Tuple[int] or None
            The input shape of the model.

        """
        if self.input_type == "waveform":
            input_shape = tuple(waveform_shape)
        elif self.input_type == "fft":
            nfft = self.nfft or waveform_shape[-1]
            seq_len = torch.fft.rfftfreq(nfft).shape[0]
            if self.drop_dc:
                seq_len -= 1
            input_shape = (*waveform_shape[:-2], 2 * waveform_shape[-2], seq_len)
        elif self.input_type == "spectrogram":
            n_samples = compute_conv_output_shape(
                waveform_shape
                if len(waveform_shape) == 3
                else [None] + list(waveform_shape),
                kernel_size=self.win_length,
                stride=self.hop_length,
                asymmetric_padding=[self.hop_length, self.win_length - self.hop_length],
            )[-1]
            if self.feature_fs is not None:
                n_samples = math.floor(n_samples * self.feature_fs / self.fs)
            if self.to1d:
                mid_dims = (self.n_channels * self.n_bins,)
            else:
                mid_dims = (self.n_channels, self.n_bins)
            input_shape = (*waveform_shape[:-2], *mid_dims, n_samples)

        if len(waveform_shape) == 2 and self._config.ensure_batch_dim:
            input_shape = (1, *input_shape)

        return input_shape

    def extra_repr_keys(self) -> List[str]:
        return ["input_type", "n_channels", "n_samples", "dtype", "device"]


class WaveformInput(BaseInput):
    """Waveform input.

    Examples
    --------
    >>> from torch_ecg.cfg import DEFAULTS
    >>> BATCH_SIZE = 32
    >>> N_CHANNELS = 12
    >>> N_SAMPLES = 5000
    >>> input_config = InputConfig(
    ...     input_type="waveform",
    ...     n_channels=N_CHANNELS,
    ...     n_samples=N_SAMPLES,
    ... )
    >>> inputer = WaveformInput(input_config)
    >>> waveform = torch.randn(BATCH_SIZE, N_CHANNELS, N_SAMPLES)
    >>> inputer(waveform).shape
    torch.Size([32, 12, 5000])
    >>> waveform = DEFAULTS.RNG.uniform(size=(N_CHANNELS, N_SAMPLES))
    >>> inputer(waveform).shape
    torch.Size([1, 12, 5000])
    >>> input_config = InputConfig(
    ...     input_type="waveform",
    ...     n_channels=N_CHANNELS,
    ...     n_samples=N_SAMPLES,
    ...     ensure_batch_dim=False,
    ... )
    >>> inputer = WaveformInput(input_config)
    >>> waveform = DEFAULTS.RNG.uniform(size=(N_CHANNELS, N_SAMPLES))
    >>> inputer(waveform).shape
    torch.Size([12, 5000])

    """

    __name__ = "WaveformInput"

    def _post_init(self) -> None:
        """Make sure the input type is `waveform`."""
        assert self.input_type == "waveform", "`input_type` must be `waveform`"

    def _from_waveform(self, waveform: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        """Internal method to convert the waveform to the input tensor."""
        self._values = torch.as_tensor(waveform).to(self.device, self.dtype)
        return self._values

    def from_waveform(self, waveform: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        """Converts the input :class:`~numpy.ndarray` or
        :class:`~torch.Tensor` waveform to a :class:`~torch.Tensor`.

        Parameters
        ----------
        waveform : numpy.ndarray or torch.Tensor
            The waveform to be transformed,
            of shape ``(batch_size, n_channels, n_samples)``
            or ``(n_channels, n_samples)``.

        Returns
        -------
        torch.Tensor
            The transformed waveform,
            of shape ``(batch_size, n_channels, n_samples)``.

        NOTE
        ----
        If the input is a 2D tensor,
        then the batch dimension is added (batch_size = 1).

        """
        return super().from_waveform(waveform)

    @add_docstring(from_waveform.__doc__)
    def __call__(self, waveform: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        """ """
        return self.from_waveform(waveform)


class FFTInput(BaseInput):
    """Inputs from the FFT, via concatenating the amplitudes and the phases.

    One can set the following optional parameters for initialization:
        - nfft: int
            the number of FFT bins.
            If nfft is None, the number of FFT bins is computed from the input shape.
        - drop_dc: bool, default True
            Whether to drop the zero frequency bin (the DC component).
        - norm: str, optional
            The normalization of the FFT, can be
                - "forward"
                - "backward"
                - "ortho"

    Examples
    --------
    >>> from torch_ecg.cfg import DEFAULTS
    >>> BATCH_SIZE = 32
    >>> N_CHANNELS = 12
    >>> N_SAMPLES = 5000
    >>> input_config = InputConfig(
    ...     input_type="fft",
    ...     n_channels=N_CHANNELS,
    ...     n_samples=N_SAMPLES,
    ...     n_fft=200,
    ...     drop_dc=True,
    ...     norm="ortho",
    ... )
    >>> inputer = FFTInput(input_config)
    >>> waveform = torch.randn(BATCH_SIZE, N_CHANNELS, N_SAMPLES)
    >>> inputer(waveform).ndim
    3
    >>> inputer(waveform).shape == inputer.compute_input_shape(waveform.shape)
    True
    >>> waveform = DEFAULTS.RNG.uniform(size=(N_CHANNELS, N_SAMPLES))
    >>> inputer(waveform).ndim
    3
    >>> inputer(waveform).shape == inputer.compute_input_shape(waveform.shape)
    True
    >>> input_config = InputConfig(
    ...     input_type="fft",
    ...     n_channels=N_CHANNELS,
    ...     n_samples=N_SAMPLES,
    ...     n_fft=None,
    ...     drop_dc=False,
    ...     norm="forward",
    ...     ensure_batch_dim=False,
    ... )
    >>> inputer = FFTInput(input_config)
    >>> waveform = DEFAULTS.RNG.uniform(size=(N_CHANNELS, N_SAMPLES))
    >>> inputer(waveform).ndim
    2
    >>> inputer(waveform).shape == inputer.compute_input_shape(waveform.shape)
    True

    """

    __name__ = "FFTInput"

    def _post_init(self) -> None:
        """Make sure the input type is `fft` and set the parameters."""
        assert self.input_type == "fft", "`input_type` must be `fft`"
        self.nfft = self._config.get("nfft", None)
        if self.nfft is None and self.n_samples > 0:
            self.nfft = self.n_samples
        self.drop_dc = self._config.get("drop_dc", True)
        self.norm = self._config.get("norm", None)
        if self.norm is not None:
            assert self.norm in [
                "forward",
                "backward",
                "ortho",
            ], f"`norm` must be one of [`forward`, `backward`, `ortho`], got {self.norm}"

    def _from_waveform(self, waveform: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        """Internal method to convert the waveform to the input tensor."""
        self._values = torch.fft.rfft(
            torch.as_tensor(waveform).to(self.device, self.dtype),
            n=self.nfft,
            dim=-1,
            norm=self.norm,
        )
        if self.drop_dc:
            self._values = self._values[..., 1:]
        self._values = torch.cat(
            [torch.abs(self._values), torch.angle(self._values)], dim=-2
        )
        return self._values

    def from_waveform(self, waveform: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        """Converts the input :class:`~numpy.ndarray` or
        :class:`~torch.Tensor` waveform to a :class:`~torch.Tensor` of FFTs.

        Parameters
        ----------
        waveform : numpy.ndarray or torch.Tensor
            The waveform to be transformed,
            of shape ``(batch_size, n_channels, n_samples)``
            or ``(n_channels, n_samples)``.

        Returns
        -------
        torch.Tensor
            The transformed waveform,
            of shape ``(batch_size, 2 * n_channels, seq_len)``,
            where `seq_len` is computed via :code:`torch.fft.rfftfreq(nfft).shape[0]`,
            if `drop_dc` is True, then seq_len is reduced by 1

        NOTE
        ----
        If the input is a 2D tensor,
        then the batch dimension is added (batch_size = 1).

        """
        return super().from_waveform(waveform)

    @add_docstring(from_waveform.__doc__)
    def __call__(self, waveform: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        return self.from_waveform(waveform)

    def extra_repr_keys(self) -> List[str]:
        return super().extra_repr_keys() + ["nfft", "drop_dc", "norm"]


class _SpectralInput(BaseInput):
    """Inputs from the spectro-temporal domain.

    One has to set the following parameters for initialization:
        - n_bins : int
            The number of frequency bins.
        - fs (or sample_rate) : int
            The sample rate of the waveform.
    with the following optional parameters with default values:
        - window_size : float, default: 1 / 20
            The size of the window in seconds.
        - overlap_size : float, default: 1 / 40
            The overlap of the windows in seconds.
        - feature_fs : None or float,
            The sample rate of the features.
            If specified, the features will be resampled
            against `fs` to this sample rate.
        - to1d : bool, default False
            Whether to convert the features to 1D.
            NOTE that if `to1d` is True,
            then if the convolutions with ``groups=1`` applied to the `input`
            acts on all the bins, which is "global"
            w.r.t. the `bins` dimension of the corresponding 2d input.

    """

    __name__ = "_SpectralInput"

    def _post_init(self) -> None:
        """Make sure the input type is `spectral` and set the parameters."""
        self.to1d = self._config.get("to1d", False)
        self.fs = self._config.get("fs", self._config.get("sample_rate"))
        self.feature_fs = self._config.get("feature_fs", None)
        if "window_size" not in self._config:
            self._config.window_size = 1 / 20
        assert (
            0 < self._config.window_size < 0.2
        ), f"`window_size` must be in (0, 0.2), got {self._config.window_size}"
        if "overlap_size" not in self._config:
            self._config.overlap_size = 1 / 40
        # TODO: consider negative overlap_size, i.e. positive gaps between windows
        assert 0 < self._config.overlap_size < self._config.window_size, (
            f"`overlap_size` must be in `(0, window_size)` = {(0, self._config.window_size)}, "
            f"got {self._config.overlap_size}"
        )

    @property
    def n_bins(self) -> int:
        return self._config.n_bins

    @property
    def window_size(self) -> int:
        return round(self._config.window_size * self.fs)

    @property
    def win_length(self) -> int:
        return self.window_size

    @property
    def overlap_size(self) -> int:
        return round(self._config.overlap_size * self.fs)

    @property
    def hop_length(self) -> int:
        return self.window_size - self.overlap_size

    @property
    def input_channels(self) -> int:
        channel_dim = -2 if self.to1d else -3
        if self.values is not None:
            return self.values.shape[channel_dim]
        return self.compute_input_shape((self.n_channels, self.n_samples))[channel_dim]

    @property
    def input_samples(self) -> Tuple[int, ...]:
        sample_dim = (-1,) if self.to1d else (-2, -1)
        if self.values is not None:
            input_shape = self.values.shape
        input_shape = self.compute_input_shape((self.n_channels, self.n_samples))
        return tuple(input_shape[dim] for dim in sample_dim)

    def extra_repr_keys(self) -> List[str]:
        return super().extra_repr_keys() + [
            "n_bins",
            "win_length",
            "hop_length",
            "fs",
            "feature_fs",
            "to1d",
        ]


class SpectrogramInput(_SpectralInput):

    __doc__ = (
        _SpectralInput.__doc__
        + """

    Examples
    --------
    >>> from torch_ecg.cfg import DEFAULTS
    >>> BATCH_SIZE = 32
    >>> N_CHANNELS = 12
    >>> N_SAMPLES = 5000
    >>> input_config = InputConfig(
    ...     name="spectrogram",
    ...     n_channels=N_CHANNELS,
    ...     n_samples=N_SAMPLES,
    ...     n_bins=128,
    ...     fs=500,
    ...     window_size=1 / 20,
    ...     overlap_size=1 / 40,
    ...     feature_fs=100,
    ...     to1d=True,
    ... )
    >>> inputer = SpectrogramInput(input_config)
    >>> waveform = torch.randn(BATCH_SIZE, N_CHANNELS, N_SAMPLES)
    >>> spectrogram = inputer(waveform)
    >>> spectrogram.shape == inputer.compute_input_shape(waveform.shape)
    True

    """
    )
    __name__ = "SpectrogramInput"

    def _post_init(self) -> None:
        """Make sure the input type is `spectrogram` and set the parameters."""
        super()._post_init()
        assert self.input_type in [
            "spectrogram"
        ], f"`input_type` must be one of [`spectrogram`], got {self.input_type}"
        args = inspect.getfullargspec(Spectrogram.__init__).args
        for k in ["self", "n_fft", "win_length", "hop_length"]:
            args.remove(k)
        kwargs = {k: self._config[k] for k in args if k in self._config}
        kwargs["n_fft"] = (self.n_bins - 1) * 2
        kwargs["win_length"] = self.win_length
        kwargs["hop_length"] = self.hop_length
        self._transform = torch.nn.Sequential()
        self._transform.add_module(
            "spectrogram", Spectrogram(**kwargs).to(self.device, self.dtype)
        )
        if self.to1d:
            self._transform.add_module(
                "to1d",
                Rearrange("... channel n_bins time -> ... (channel n_bins) time"),
            )

    def _from_waveform(self, waveform: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        """Internal method to convert the waveform to the input tensor."""
        self._values = self._transform(
            torch.as_tensor(waveform).to(self.device, self.dtype)
        )
        if self.feature_fs is not None:
            # self.values.ndim can be 2, 3, or 4
            scale_factor = [1] * (self.values.ndim - 3) + [self.feature_fs / self.fs]
            if self.values.ndim == 2:
                self._values = F.interpolate(
                    self._values.unsqueeze(0),
                    scale_factor=scale_factor,
                    recompute_scale_factor=True,
                ).squeeze(0)
            else:
                self._values = F.interpolate(
                    self._values, scale_factor=scale_factor, recompute_scale_factor=True
                )
        return self._values

    def from_waveform(self, waveform: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        r"""Converts the input :class:`~numpy.ndarray` or
        :class:`~torch.Tensor` waveform to a :class:`~torch.Tensor` of spectrograms.

        Parameters
        ----------
        waveform : numpy.ndarray or torch.Tensor
            The waveform to be transformed,
            of shape ``(batch_size, n_channels, n_samples)``
            or ``(n_channels, n_samples)``.

        Returns
        -------
        torch.Tensor
            The transformed waveform,
            of shape ``(batch_size, n_channels, n_bins, n_frames)``, where

            .. math::

                n\_frames = (n\_samples - win\_length) // hop\_length + 1

        NOTE
        ----
        If the input is a 2D tensor,
        then the batch dimension is added (batch_size = 1).

        """
        return super().from_waveform(waveform)

    @add_docstring(from_waveform.__doc__)
    def __call__(self, waveform: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        return self.from_waveform(waveform)
